# stdlib
from collections.abc import Callable

# third party
import requests

# relative
from ...serde.serializable import serializable
from ...store.document_store import DocumentStore
from ...util.telemetry import instrument
from ..context import AuthedServiceContext
from ..network.routes import VeilidNodeRoute
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import service_method
from ..user.user_roles import DATA_OWNER_ROLE_LEVEL
from .veilid_endpoints import GEN_VLD_KEY_ENDPOINT
from .veilid_endpoints import HEALTHCHECK_ENDPOINT
from .veilid_endpoints import RET_VLD_KEY_ENDPOINT
from .veilid_endpoints import VEILID_SERVICE_URL


@instrument
@serializable()
class VeilidService(AbstractService):
    store: DocumentStore

    def __init__(self, store: DocumentStore) -> None:
        self.store = store

    def perform_request(
        self, method: Callable, endpoint: str, raw: bool = False
    ) -> SyftSuccess | SyftError | str:
        try:
            response = method(f"{VEILID_SERVICE_URL}{endpoint}")
            response.raise_for_status()
            message = response.json().get("message")
            return message if raw else SyftSuccess(message=message)
        except requests.HTTPError:
            return SyftError(message=f"{response.json()['detail']}")
        except requests.RequestException as e:
            return SyftError(message=f"Failed to perform request. {e}")

    def is_veilid_service_healthy(self) -> bool:
        res = self.perform_request(
            method=requests.get, endpoint=HEALTHCHECK_ENDPOINT, raw=True
        )
        return res == "OK"

    @service_method(
        path="veilid.generate_vld_key",
        name="generate_vld_key",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def generate_vld_key(self, context: AuthedServiceContext) -> str | SyftError:
        if not self.is_veilid_service_healthy():
            return SyftError(
                message="Veilid service is not healthy. Please try again later."
            )
        return self.perform_request(
            method=requests.post,
            endpoint=GEN_VLD_KEY_ENDPOINT,
        )

    @service_method(
        path="veilid.retrieve_vld_key",
        name="retrieve_vld_key",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def retrieve_vld_key(self, context: AuthedServiceContext) -> str | SyftError:
        if not self.is_veilid_service_healthy():
            return SyftError(
                message="Veilid service is not healthy. Please try again later."
            )
        return self.perform_request(
            method=requests.get,
            endpoint=RET_VLD_KEY_ENDPOINT,
            raw=True,
        )

    @service_method(
        path="veilid.get_veilid_route",
        name="get_veilid_route",
    )
    def get_veilid_route(
        self, context: AuthedServiceContext
    ) -> VeilidNodeRoute | SyftError:
        vld_key = self.retrieve_vld_key(context)
        if isinstance(vld_key, SyftError):
            return vld_key
        return VeilidNodeRoute(vld_key=vld_key)



# stdlib
from collections import defaultdict
from typing import Any
from typing import cast

# third party
from result import Result

# relative
from ...abstract_node import AbstractNode
from ...client.api import NodeIdentity
from ...node.credentials import SyftVerifyKey
from ...serde.serializable import serializable
from ...store.document_store import BaseStash
from ...store.document_store import DocumentStore
from ...store.linked_obj import LinkedObject
from ...types.syft_object import SyftObject
from ...types.syncable_object import SyncableSyftObject
from ...types.uid import UID
from ...util.telemetry import instrument
from ..action.action_object import ActionObject
from ..action.action_permissions import ActionObjectPermission
from ..action.action_permissions import ActionPermission
from ..action.action_permissions import StoragePermission
from ..code.user_code import UserCodeStatusCollection
from ..context import AuthedServiceContext
from ..job.job_stash import Job
from ..output.output_service import ExecutionOutput
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user_roles import ADMIN_ROLE_LEVEL
from .sync_stash import SyncStash
from .sync_state import SyncState


def get_store(context: AuthedServiceContext, item: SyncableSyftObject) -> Any:
    if isinstance(item, ActionObject):
        service = context.node.get_service("actionservice")  # type: ignore
        return service.store  # type: ignore
    service = context.node.get_service(TYPE_TO_SERVICE[type(item)])  # type: ignore
    return service.stash.partition


@instrument
@serializable()
class SyncService(AbstractService):
    store: DocumentStore
    stash: SyncStash

    def __init__(self, store: DocumentStore):
        self.store = store
        self.stash = SyncStash(store=store)

    def add_actionobject_read_permissions(
        self,
        context: AuthedServiceContext,
        action_object: ActionObject,
        permissions_other: list[str],
    ) -> None:
        read_permissions = [x for x in permissions_other if "READ" in x]

        _id = action_object.id.id
        blob_id = action_object.syft_blob_storage_entry_id

        store_to = context.node.get_service("actionservice").store  # type: ignore
        store_to_blob = context.node.get_service("blobstorageservice").stash.partition  # type: ignore

        for read_permission in read_permissions:
            creds, perm_str = read_permission.split("_")
            perm = ActionPermission[perm_str]
            permission = ActionObjectPermission(
                uid=_id, permission=perm, credentials=SyftVerifyKey(creds)
            )
            store_to.add_permission(permission)

            permission_blob = ActionObjectPermission(
                uid=blob_id, permission=perm, credentials=SyftVerifyKey(creds)
            )
            store_to_blob.add_permission(permission_blob)

    def set_obj_ids(self, context: AuthedServiceContext, x: Any) -> None:
        if hasattr(x, "__dict__") and isinstance(x, SyftObject):
            for val in x.__dict__.values():
                if isinstance(val, list | tuple):
                    for v in val:
                        self.set_obj_ids(context, v)
                elif isinstance(val, dict):
                    for v in val.values():
                        self.set_obj_ids(context, v)
                else:
                    self.set_obj_ids(context, val)
            x.syft_node_location = context.node.id  # type: ignore
            x.syft_client_verify_key = context.credentials
            if hasattr(x, "node_uid"):
                x.node_uid = context.node.id  # type: ignore

    def transform_item(
        self, context: AuthedServiceContext, item: SyftObject
    ) -> SyftObject:
        if isinstance(item, UserCodeStatusCollection):
            identity = NodeIdentity.from_node(context.node)
            res = {}
            for key in item.status_dict.keys():
                # todo, check if they are actually only two nodes
                res[identity] = item.status_dict[key]
            item.status_dict = res

        self.set_obj_ids(context, item)
        return item

    def get_stash_for_item(
        self, context: AuthedServiceContext, item: SyftObject
    ) -> BaseStash:
        services = list(context.node.service_path_map.values())  # type: ignore

        all_stashes = {}
        for serv in services:
            if (_stash := getattr(serv, "stash", None)) is not None:
                all_stashes[_stash.object_type] = _stash

        stash = all_stashes.get(type(item), None)
        return stash

    def add_permissions_for_item(
        self,
        context: AuthedServiceContext,
        item: SyftObject,
        permissions_other: set[ActionObjectPermission],
    ) -> None:
        if isinstance(item, Job) and context.node.node_side_type.value == "low":  # type: ignore
            _id = item.id
            read_permissions = [x for x in permissions_other if "READ" in x]  # type: ignore
            job_store = context.node.get_service("jobservice").stash.partition  # type: ignore
            for read_permission in read_permissions:
                creds, perm_str = read_permission.split("_")
                perm = ActionPermission[perm_str]
                permission = ActionObjectPermission(
                    uid=_id, permission=perm, credentials=SyftVerifyKey(creds)
                )
                job_store.add_permission(permission)

    def add_storage_permissions_for_item(
        self,
        context: AuthedServiceContext,
        item: SyftObject,
        permissions_other: set[UID],
    ) -> None:
        _id = item.id.id
        permissions = [
            StoragePermission(uid=_id, node_uid=p) for p in permissions_other
        ]

        store = get_store(context, item)
        store.add_storage_permissions(permissions)

    def set_object(
        self, context: AuthedServiceContext, item: SyncableSyftObject
    ) -> Result[SyftObject, str]:
        stash = self.get_stash_for_item(context, item)
        creds = context.credentials

        exists = stash.get_by_uid(context.credentials, item.id).ok() is not None
        if exists:
            res = stash.update(creds, item)
        else:
            # Storage permissions are added separately
            res = stash.set(
                creds,
                item,
                add_storage_permission=False,
            )

        return res

    @service_method(
        path="sync.sync_items",
        name="sync_items",
        roles=ADMIN_ROLE_LEVEL,
    )
    def sync_items(
        self,
        context: AuthedServiceContext,
        items: list[ActionObject | SyftObject],
        permissions: dict[UID, set[str]],
        storage_permissions: dict[UID, set[UID]],
    ) -> SyftSuccess | SyftError:
        permissions = defaultdict(set, permissions)
        storage_permissions = defaultdict(set, storage_permissions)
        for item in items:
            new_permissions = permissions[item.id.id]
            new_storage_permissions = storage_permissions[item.id.id]
            if isinstance(item, ActionObject):
                self.add_actionobject_read_permissions(context, item, new_permissions)
                self.add_storage_permissions_for_item(
                    context, item, new_storage_permissions
                )
            else:
                item = self.transform_item(context, item)  # type: ignore[unreachable]
                res = self.set_object(context, item)

                if res.is_ok():
                    self.add_permissions_for_item(context, item, new_permissions)
                    self.add_storage_permissions_for_item(
                        context, item, new_storage_permissions
                    )
                else:
                    return SyftError(message=f"Failed to sync {res.err()}")
        return SyftSuccess(message=f"Synced {len(items)} items")

    @service_method(
        path="sync.get_permissions",
        name="get_permissions",
        roles=ADMIN_ROLE_LEVEL,
    )
    def get_permissions(
        self,
        context: AuthedServiceContext,
        items: list[SyncableSyftObject],
    ) -> tuple[dict[UID, set[str]], dict[UID, set[str]]]:
        permissions = {}
        storage_permissions = {}

        for item in items:
            store = get_store(context, item)
            if store is not None:
                _id = item.id.id
                permissions[_id] = store.permissions[_id]
                storage_permissions[_id] = store.storage_permissions[_id]
        return permissions, storage_permissions

    @service_method(
        path="sync._get_state",
        name="_get_state",
        roles=ADMIN_ROLE_LEVEL,
    )
    def _get_state(
        self, context: AuthedServiceContext, add_to_store: bool = False
    ) -> SyncState | SyftError:
        node = cast(AbstractNode, context.node)

        new_state = SyncState(node_uid=node.id)

        services_to_sync = [
            "requestservice",
            "usercodeservice",
            "jobservice",
            "logservice",
            "outputservice",
            "usercodestatusservice",
        ]

        for service_name in services_to_sync:
            service = node.get_service(service_name)
            items = service.get_all(context)
            new_state.add_objects(items, api=node.root_client.api)  # type: ignore

        # TODO workaround, we only need action objects from outputs for now
        action_object_ids = set()
        for obj in new_state.objects.values():
            if isinstance(obj, ExecutionOutput):
                action_object_ids |= set(obj.output_id_list)
            elif isinstance(obj, Job) and obj.result is not None:
                if isinstance(obj.result, ActionObject):
                    obj.result = obj.result.as_empty()
                action_object_ids.add(obj.result.id)

        action_objects = []
        for uid in action_object_ids:
            action_object = node.get_service("actionservice").get(context, uid)  # type: ignore
            if action_object.is_err():
                return SyftError(message=action_object.err())
            action_objects.append(action_object.ok())
        new_state.add_objects(action_objects)

        new_state._build_dependencies(api=node.root_client.api)  # type: ignore

        permissions, storage_permissions = self.get_permissions(
            context, new_state.objects.values()
        )
        new_state.permissions = permissions
        new_state.storage_permissions = storage_permissions

        previous_state = self.stash.get_latest(context=context)
        if previous_state is not None:
            new_state.previous_state_link = LinkedObject.from_obj(
                obj=previous_state,
                service_type=SyncService,
                node_uid=context.node.id,  # type: ignore
            )

        if add_to_store:
            self.stash.set(context.credentials, new_state)

        return new_state



"""
How to check differences between two objects:
    * by default merge every attr
    * check if there is a custom implementation of the check function
    * check if there are exceptions we do not want to merge
    * check if there are some restrictions on the attr set
"""

# stdlib
import html
import textwrap
from typing import Any
from typing import ClassVar

# third party
from pydantic import model_validator
from rich import box
from rich.console import Console
from rich.console import Group
from rich.markdown import Markdown
from rich.padding import Padding
from rich.panel import Panel
from typing_extensions import Self

# relative
from ...types.syft_object import SYFT_OBJECT_VERSION_2
from ...types.syft_object import SyftObject
from ...types.syncable_object import SyncableSyftObject
from ...types.uid import LineageID
from ...types.uid import UID
from ...util import options
from ...util.colors import SURFACE
from ...util.fonts import ITABLES_CSS
from ...util.fonts import fonts_css
from ..action.action_object import ActionObject
from ..action.action_permissions import ActionObjectPermission
from ..action.action_permissions import StoragePermission
from ..code.user_code import UserCode
from ..code.user_code import UserCodeStatusCollection
from ..job.job_stash import Job
from ..log.log import SyftLog
from ..output.output_service import ExecutionOutput
from ..request.request import Request
from ..response import SyftError
from .sync_state import SyncState

sketchy_tab = "‎ " * 4


class AttrDiff(SyftObject):
    # version
    __canonical_name__ = "AttrDiff"
    __version__ = SYFT_OBJECT_VERSION_2
    attr_name: str
    low_attr: Any = None
    high_attr: Any = None

    def _repr_html_(self) -> str:
        return f"""{self.attr_name}:
    Low Side value: {self.low_attr}
    High Side value: {self.high_attr}
    """

    def __repr_side__(self, side: str) -> str:
        if side == "low":
            return recursive_attr_repr(self.low_attr)
        else:
            return recursive_attr_repr(self.high_attr)

    def _coll_repr_(self) -> dict[str, Any]:
        return {
            "attr name": self.attr_name,
            "low attr": html.escape(f"{self.low_attr}"),
            "high attr": html.escape(str(self.high_attr)),
        }


class ListDiff(AttrDiff):
    # version
    __canonical_name__ = "ListDiff"
    __version__ = SYFT_OBJECT_VERSION_2
    diff_ids: list[int] = []
    new_low_ids: list[int] = []
    new_high_ids: list[int] = []

    @property
    def is_empty(self) -> bool:
        return (
            len(self.diff_ids) == 0
            and len(self.new_low_ids) == 0
            and len(self.new_high_ids) == 0
        )

    @classmethod
    def from_lists(cls, attr_name: str, low_list: list, high_list: list) -> "ListDiff":
        diff_ids = []
        new_low_ids = []
        new_high_ids = []
        if len(low_list) != len(high_list):
            if len(low_list) > len(high_list):
                common_length = len(high_list)
                new_low_ids = list(range(common_length, len(low_list)))
            else:
                common_length = len(low_list)
                new_high_ids = list(range(common_length, len(high_list)))
        else:
            common_length = len(low_list)

        for i in range(common_length):
            # if hasattr(low_list[i], 'syft_eq'):
            #     if not low_list[i].syft_eq(high_list[i]):
            #         diff_ids.append(i)
            if low_list[i] != high_list[i]:
                diff_ids.append(i)

        change_diff = ListDiff(
            attr_name=attr_name,
            low_attr=low_list,
            high_attr=high_list,
            diff_ids=diff_ids,
            new_low_ids=new_low_ids,
            new_high_ids=new_high_ids,
        )
        return change_diff


def recursive_attr_repr(value_attr: list | dict | bytes, num_tabs: int = 0) -> str:
    new_num_tabs = num_tabs + 1

    if isinstance(value_attr, list):
        list_repr = "[\n"
        for elem in value_attr:
            list_repr += recursive_attr_repr(elem, num_tabs=num_tabs + 1) + "\n"
        list_repr += "]"
        return list_repr

    elif isinstance(value_attr, dict):
        dict_repr = "{\n"
        for key, elem in value_attr.items():
            dict_repr += f"{sketchy_tab * new_num_tabs}{key}: {str(elem)}\n"
        dict_repr += "}"
        return dict_repr

    elif isinstance(value_attr, bytes):
        value_attr = repr(value_attr)  # type: ignore
        if len(value_attr) > 50:
            value_attr = value_attr[:50] + "..."  # type: ignore
    return f"{sketchy_tab*num_tabs}{str(value_attr)}"


class ObjectDiff(SyftObject):  # StateTuple (compare 2 objects)
    # version
    __canonical_name__ = "ObjectDiff"
    __version__ = SYFT_OBJECT_VERSION_2
    low_obj: SyncableSyftObject | None = None
    high_obj: SyncableSyftObject | None = None
    low_node_uid: UID
    high_node_uid: UID
    low_permissions: list[str] = []
    high_permissions: list[str] = []
    low_storage_permissions: set[UID] = set()
    high_storage_permissions: set[UID] = set()

    obj_type: type
    diff_list: list[AttrDiff] = []

    __repr_attrs__ = [
        "low_state",
        "high_state",
    ]

    def is_mock(self, side: str) -> bool:
        # An object is a mock object if it exists on both sides,
        # and has no storage permissions on `side`
        # NOTE both sides must have the objects, else it is a new object.
        # New+mock objects do not appear naturally, but if they do we
        # want them to show up.
        if side == "low":
            obj = self.low_obj
            other_obj = self.high_obj
            permissions = self.low_storage_permissions
            node_uid = self.low_node_uid
        elif side == "high":
            obj = self.high_obj
            other_obj = self.low_obj
            permissions = self.high_storage_permissions
            node_uid = self.high_node_uid
        else:
            raise ValueError("Invalid side")

        if obj is None or other_obj is None:
            return False

        return node_uid not in permissions

    @classmethod
    def from_objects(
        cls,
        low_obj: SyncableSyftObject | None,
        high_obj: SyncableSyftObject | None,
        low_permissions: set[str],
        high_permissions: set[str],
        low_storage_permissions: set[UID],
        high_storage_permissions: set[UID],
        low_node_uid: UID,
        high_node_uid: UID,
    ) -> "ObjectDiff":
        if low_obj is None and high_obj is None:
            raise ValueError("Both low and high objects are None")
        obj_type = type(low_obj if low_obj is not None else high_obj)

        res = cls(
            low_obj=low_obj,
            high_obj=high_obj,
            obj_type=obj_type,
            low_node_uid=low_node_uid,
            high_node_uid=high_node_uid,
            low_permissions=low_permissions,
            high_permissions=high_permissions,
            low_storage_permissions=low_storage_permissions,
            high_storage_permissions=high_storage_permissions,
        )

        if (
            low_obj is None
            or high_obj is None
            or res.is_mock("low")
            or res.is_mock("high")
        ):
            diff_list = []
        else:
            diff_list = low_obj.syft_get_diffs(high_obj)

        res.diff_list = diff_list
        return res

    def __hash__(self) -> int:
        return hash(self.id) + hash(self.low_obj) + hash(self.high_obj)

    @property
    def status(self) -> str:
        if self.low_obj is None or self.high_obj is None:
            return "NEW"
        if len(self.diff_list) == 0:
            return "SAME"
        return "DIFF"

    @property
    def object_id(self) -> UID:
        uid: UID | LineageID = (
            self.low_obj.id if self.low_obj is not None else self.high_obj.id  # type: ignore
        )
        if isinstance(uid, LineageID):
            return uid.id
        return uid

    @property
    def non_empty_object(self) -> SyftObject | None:
        return self.low_obj or self.high_obj

    @property
    def object_type(self) -> str:
        return self.obj_type.__name__

    @property
    def high_state(self) -> str:
        return self.state_str("high")

    @property
    def low_state(self) -> str:
        return self.state_str("low")

    @property
    def object_uid(self) -> UID:
        return self.low_obj.id if self.low_obj is not None else self.high_obj.id  # type: ignore

    def diff_attributes_str(self, side: str) -> str:
        obj = self.low_obj if side == "low" else self.high_obj

        if obj is None:
            return ""

        repr_attrs = getattr(obj, "__repr_attrs__", [])
        if self.status == "SAME":
            repr_attrs = repr_attrs[:3]

        if self.status in {"SAME", "NEW"}:
            attrs_str = ""
            for attr in repr_attrs:
                value = getattr(obj, attr)
                attrs_str += f"{attr}: {recursive_attr_repr(value)}\n"
            return attrs_str

        elif self.status == "DIFF":
            attrs_str = ""
            for diff in self.diff_list:
                attrs_str += f"{diff.attr_name}: {diff.__repr_side__(side)}\n"
            return attrs_str
        else:
            raise ValueError("")

    def diff_side_str(self, side: str) -> str:
        obj = self.low_obj if side == "low" else self.high_obj
        if obj is None:
            return ""
        res = f"{self.obj_type.__name__.upper()} #{obj.id}:\n"
        res += self.diff_attributes_str(side)
        return res

    def state_str(self, side: str) -> str:
        other_obj: SyftObject | None = None
        if side == "high":
            obj = self.high_obj
            other_obj = self.low_obj
        else:
            obj = self.low_obj
            other_obj = self.high_obj

        if obj is None:
            return "-"
        if self.status == "SAME":
            return f"SAME\n{self.obj_type.__name__}"

        if isinstance(obj, ActionObject):
            return obj.__repr__()

        if other_obj is None:  # type: ignore[unreachable]
            attrs_str = ""
            attrs = getattr(obj, "__repr_attrs__", [])
            for attr in attrs:
                value = getattr(obj, attr)
                attrs_str += f"{sketchy_tab}{attr} = {recursive_attr_repr(value)}\n"
            attrs_str = attrs_str[:-1]
            return f"NEW\n\nclass {self.object_type}:\n{attrs_str}"

        attr_text = f"DIFF\nclass {self.object_type}:\n"
        for diff in self.diff_list:
            # TODO
            attr_text += (
                f"{sketchy_tab}{diff.attr_name}={diff.__repr_side__(side)}," + "\n"
            )
        if len(self.diff_list) > 0:
            attr_text = attr_text[:-2]

        return attr_text

    def get_obj(self) -> SyftObject | None:
        if self.status == "NEW":
            return self.low_obj if self.low_obj is not None else self.high_obj
        else:
            raise ValueError("ERROR")

    def _coll_repr_(self) -> dict[str, Any]:
        low_state = f"{self.status}\n{self.diff_side_str('low')}"
        high_state = f"{self.status}\n{self.diff_side_str('high')}"
        return {
            "low_state": html.escape(low_state),
            "high_state": html.escape(high_state),
        }

    def _repr_html_(self) -> str:
        if self.low_obj is None and self.high_obj is None:
            return SyftError(message="Something broke")

        base_str = f"""
        <style>
        {fonts_css}
        .syft-dataset {{color: {SURFACE[options.color_theme]};}}
        .syft-dataset h3,
        .syft-dataset p
            {{font-family: 'Open Sans';}}
            {ITABLES_CSS}
        </style>
        <div class='syft-diff'>
        """

        obj_repr: str
        attr_text: str
        if self.low_obj is None:
            if hasattr(self.high_obj, "_repr_html_"):
                obj_repr = self.high_obj._repr_html_()  # type: ignore
            elif hasattr(self.high_obj, "_inner_repr"):
                obj_repr = self.high_obj._inner_repr()  # type: ignore
            else:
                obj_repr = self.__repr__()
            attr_text = (
                f"""
    <h3>{self.object_type} ObjectDiff (New {self.object_type}  on the High Side):</h3>
    """
                + obj_repr
            )

        elif self.high_obj is None:
            if hasattr(self.low_obj, "_repr_html_"):
                obj_repr = self.low_obj._repr_html_()  # type: ignore
            elif hasattr(self.low_obj, "_inner_repr"):
                obj_repr = self.low_obj._inner_repr()  # type: ignore
            else:
                obj_repr = self.__repr__()
            attr_text = (
                f"""
    <h3>{self.object_type} ObjectDiff (New {self.object_type}  on the High Side):</h3>
    """
                + obj_repr
            )

        elif self.status == "SAME":
            obj_repr = "No changes between low side and high side"
        else:
            obj_repr = ""
            for diff in self.diff_list:
                obj_repr += diff.__repr__() + "<br>"

            obj_repr = obj_repr.replace("\n", "<br>")
            # print("New lines", res)

        attr_text = f"<h3>{self.object_type} ObjectDiff:</h3>\n{obj_repr}"
        return base_str + attr_text


def _wrap_text(text: str, width: int, indent: int = 4) -> str:
    """Wrap text, preserving existing line breaks"""
    return "\n".join(
        [
            "\n".join(
                textwrap.wrap(
                    line,
                    width,
                    break_long_words=False,
                    replace_whitespace=False,
                    subsequent_indent=" " * indent,
                )
            )
            for line in text.splitlines()
            if line.strip() != ""
        ]
    )


class ObjectDiffBatch(SyftObject):
    __canonical_name__ = "DiffHierarchy"
    __version__ = SYFT_OBJECT_VERSION_2
    LINE_LENGTH: ClassVar[int] = 100
    INDENT: ClassVar[int] = 4
    ORDER: ClassVar[dict] = {"low": 0, "high": 1}

    # Diffs are ordered in depth-first order,
    # the first diff is the root of the hierarchy
    diffs: list[ObjectDiff]
    hierarchy_levels: list[int]
    dependencies: dict[UID, list[UID]] = {}
    dependents: dict[UID, list[UID]] = {}

    @property
    def visual_hierarchy(self) -> tuple[type, dict]:
        # Returns
        root_obj: Request | UserCodeStatusCollection | ExecutionOutput | Any = (
            self.root.low_obj if self.root.low_obj is not None else self.root.high_obj
        )
        if isinstance(root_obj, Request):
            return Request, {
                Request: [UserCode],
                UserCode: [UserCode],
            }
        if isinstance(root_obj, UserCodeStatusCollection):
            return UserCode, {
                UserCode: [UserCodeStatusCollection],
            }
        if isinstance(root_obj, ExecutionOutput):
            return UserCode, {
                UserCode: [Job],
                Job: [ExecutionOutput, SyftLog, Job],
                ExecutionOutput: [ActionObject],
            }
        raise ValueError(f"Unknown root type: {self.root.obj_type}")

    @model_validator(mode="after")
    def make_dependents(self) -> Self:
        dependents: dict = {}
        for parent, children in self.dependencies.items():
            for child in children:
                dependents[child] = dependents.get(child, []) + [parent]
        self.dependents = dependents
        return self

    @property
    def root(self) -> ObjectDiff:
        return self.diffs[0]

    def __len__(self) -> int:
        return len(self.diffs)

    def __repr__(self) -> str:
        return f"""{self.hierarchy_str('low')}

{self.hierarchy_str('high')}
"""

    def _repr_markdown_(self, wrap_as_python: bool = True, indent: int = 0) -> str:
        return ""  # Turns off the _repr_markdown_ of SyftObject

    def _get_visual_hierarchy(self, node: ObjectDiff) -> dict[ObjectDiff, dict]:
        _, child_types_map = self.visual_hierarchy
        child_types = child_types_map.get(node.obj_type, [])
        dep_ids = self.dependencies.get(node.object_id, []) + self.dependents.get(
            node.object_id, []
        )

        result = {}
        for child_type in child_types:
            children = [
                n
                for n in self.diffs
                if n.object_id in dep_ids
                and isinstance(n.low_obj or n.high_obj, child_type)
            ]
            for child in children:
                result[child] = self._get_visual_hierarchy(child)

        return result

    def get_visual_hierarchy(self) -> "ObjectDiffBatch":
        visual_root_type = self.visual_hierarchy[0]
        # First diff with a visual root type is the visual root
        # because diffs are in depth-first order
        visual_root = [
            diff
            for diff in self.diffs
            if isinstance(diff.low_obj or diff.high_obj, visual_root_type)
        ][0]
        return {visual_root: self._get_visual_hierarchy(visual_root)}  # type: ignore

    def _get_obj_str(self, diff_obj: ObjectDiff, level: int, side: str) -> str:
        obj = diff_obj.low_obj if side == "low" else diff_obj.high_obj
        if obj is None:
            return ""
        indent = " " * level * self.INDENT
        obj_str = diff_obj.diff_side_str(side)
        obj_str = _wrap_text(obj_str, width=self.LINE_LENGTH - len(indent))

        line_prefix = indent + f"―――― {diff_obj.status} "
        line = "―" * (self.LINE_LENGTH - len(line_prefix))
        return f"""{line_prefix}{line}

{textwrap.indent(obj_str, indent)}

"""

    def hierarchy_str(self, side: str) -> str:
        def _hierarchy_str_recursive(tree: dict, level: int) -> str:
            result = ""
            for node, children in tree.items():
                result += self._get_obj_str(node, level, side)
                result += _hierarchy_str_recursive(children, level + 1)
            return result

        visual_hierarchy = self.get_visual_hierarchy()
        res = _hierarchy_str_recursive(visual_hierarchy, 0)
        if res == "":
            res = f"No {side} side changes."
        return f"""{side.upper()} SIDE STATE:

{res}"""


class NodeDiff(SyftObject):
    __canonical_name__ = "NodeDiff"
    __version__ = SYFT_OBJECT_VERSION_2

    low_node_uid: UID
    high_node_uid: UID
    obj_uid_to_diff: dict[UID, ObjectDiff] = {}
    dependencies: dict[UID, list[UID]] = {}

    @classmethod
    def from_sync_state(
        cls: type["NodeDiff"], low_state: SyncState, high_state: SyncState
    ) -> "NodeDiff":
        obj_uid_to_diff = {}
        for obj_id in set(low_state.objects.keys()) | set(high_state.objects.keys()):
            low_obj = low_state.objects.get(obj_id, None)
            low_permissions = low_state.permissions.get(obj_id, set())
            low_storage_permissions = low_state.storage_permissions.get(obj_id, set())
            high_obj = high_state.objects.get(obj_id, None)
            high_permissions = high_state.permissions.get(obj_id, set())
            high_storage_permissions = high_state.storage_permissions.get(obj_id, set())
            diff = ObjectDiff.from_objects(
                low_obj=low_obj,
                high_obj=high_obj,
                low_permissions=low_permissions,
                high_permissions=high_permissions,
                low_storage_permissions=low_storage_permissions,
                high_storage_permissions=high_storage_permissions,
                low_node_uid=low_state.node_uid,
                high_node_uid=high_state.node_uid,
            )
            obj_uid_to_diff[diff.object_id] = diff

        node_diff = cls(
            low_node_uid=low_state.node_uid,
            high_node_uid=high_state.node_uid,
            obj_uid_to_diff=obj_uid_to_diff,
        )

        node_diff._init_dependencies(low_state, high_state)
        return node_diff

    def _init_dependencies(self, low_state: SyncState, high_state: SyncState) -> None:
        all_parents = set(low_state.dependencies.keys()) | set(
            high_state.dependencies.keys()
        )
        for parent in all_parents:
            low_deps = low_state.dependencies.get(parent, [])
            high_deps = high_state.dependencies.get(parent, [])
            self.dependencies[parent] = list(set(low_deps) | set(high_deps))

    @property
    def diffs(self) -> list[ObjectDiff]:
        diffs_depthfirst = [
            diff for hierarchy in self.hierarchies for diff in hierarchy.diffs
        ]
        # deduplicate
        diffs = []
        ids = set()
        for diff in diffs_depthfirst:
            if diff.object_id not in ids:
                diffs.append(diff)
                ids.add(diff.object_id)
        return diffs

    def _repr_html_(self) -> Any:
        return self.diffs._repr_html_()

    def _sort_hierarchies(
        self, hierarchies: list[ObjectDiffBatch]
    ) -> list[ObjectDiffBatch]:
        without_usercode = []
        grouped_by_usercode: dict[UID, list[ObjectDiffBatch]] = {}
        for hierarchy in hierarchies:
            has_usercode = False
            for diff in hierarchy.diffs:
                obj = diff.low_obj if diff.low_obj is not None else diff.high_obj
                if isinstance(obj, UserCode):
                    usercode_id = obj.id
                    if usercode_id not in grouped_by_usercode:
                        grouped_by_usercode[usercode_id] = []
                    grouped_by_usercode[usercode_id].append(hierarchy)
                    has_usercode = True
                    break
            if not has_usercode:
                without_usercode.append(hierarchy)

        # Order of hierarchies, by root object type
        hierarchy_order = [UserCodeStatusCollection, Request, ExecutionOutput]
        # Sort group by hierarchy_order, then by root object id
        for hierarchy_group in grouped_by_usercode.values():
            hierarchy_group.sort(
                key=lambda x: (
                    hierarchy_order.index(x.root.obj_type),
                    x.root.object_id,
                )
            )

        # sorted = sorted groups + without_usercode
        sorted_hierarchies = []
        for grp in grouped_by_usercode.values():
            sorted_hierarchies.extend(grp)
        sorted_hierarchies.extend(without_usercode)
        return sorted_hierarchies

    @property
    def hierarchies(self) -> list[ObjectDiffBatch]:
        # Returns a list of hierarchies, where each hierarchy is a list of tuples (ObjectDiff, level),
        # in depth-first order.

        # Each hierarchy only contains one root, at the first position
        # Example: [(Diff1, 0), (Diff2, 1), (Diff3, 2), (Diff4, 1)]
        # Diff1
        # -- Diff2
        # ---- Diff3
        # -- Diff4

        def _build_hierarchy_helper(
            uid: UID, level: int = 0, visited: set | None = None
        ) -> list:
            visited = visited if visited is not None else set()

            if uid in visited:
                return []

            result = [(uid, level)]
            visited.add(uid)
            if uid in self.dependencies:
                deps = self.dependencies[uid]
                for dep_uid in self.dependencies[uid]:
                    if dep_uid not in visited:
                        # NOTE we pass visited + deps to recursive calls, to have
                        # all objects at the highest level in the hierarchy
                        # Example:
                        # ExecutionOutput
                        # -- Job
                        # ---- Result
                        # -- Result
                        # We want to omit Job.Result, because it's already in ExecutionOutput.Result
                        result.extend(
                            _build_hierarchy_helper(
                                uid=dep_uid,
                                level=level + 1,
                                visited=visited | set(deps) - {dep_uid},
                            )
                        )
            return result

        hierarchies = []
        all_ids = set(self.obj_uid_to_diff.keys())
        child_ids = {child for deps in self.dependencies.values() for child in deps}
        # Root ids are object ids with no parents
        root_ids = list(all_ids - child_ids)

        for root_uid in root_ids:
            uid_hierarchy = _build_hierarchy_helper(root_uid)
            diffs = [self.obj_uid_to_diff[uid] for uid, _ in uid_hierarchy]
            levels = [level for _, level in uid_hierarchy]

            batch_uids = {uid for uid, _ in uid_hierarchy}
            dependencies = {
                uid: [d for d in self.dependencies.get(uid, []) if d in batch_uids]
                for uid in batch_uids
            }

            batch = ObjectDiffBatch(
                diffs=diffs,
                hierarchy_levels=levels,
                dependencies=dependencies,
            )
            hierarchies.append(batch)

        hierarchies_sorted = self._sort_hierarchies(hierarchies)
        return hierarchies_sorted

    def objs_to_sync(self) -> list[SyftObject]:
        objs: list[SyftObject] = []
        for diff in self.diffs:
            if diff.status == "NEW":
                objs.append(diff.get_obj())
        return objs


class SyncDecision(SyftObject):
    __canonical_name__ = "SyncDecision"
    __version__ = SYFT_OBJECT_VERSION_2

    diff: ObjectDiff
    decision: str | None
    new_permissions_lowside: list[ActionObjectPermission]
    new_storage_permissions_lowside: list[StoragePermission]
    new_storage_permissions_highside: list[StoragePermission]
    mockify: bool


class ResolvedSyncState(SyftObject):
    __canonical_name__ = "SyncUpdate"
    __version__ = SYFT_OBJECT_VERSION_2

    node_uid: UID
    create_objs: list[SyncableSyftObject] = []
    update_objs: list[SyncableSyftObject] = []
    delete_objs: list[SyftObject] = []
    new_permissions: list[ActionObjectPermission] = []
    new_storage_permissions: list[StoragePermission] = []
    alias: str

    def add_sync_decision(self, sync_decision: SyncDecision) -> None:
        diff = sync_decision.diff

        if diff.status == "SAME":
            return

        my_obj = diff.low_obj if self.alias == "low" else diff.high_obj
        other_obj = diff.low_obj if self.alias == "high" else diff.high_obj

        if other_obj is not None and sync_decision.mockify:
            other_obj = other_obj.create_shareable_sync_copy(mock=True)

        if sync_decision.decision != self.alias:  # chose for the other
            if diff.status == "DIFF":
                # keep IDs comparison here, otherwise it will break with actionobjects
                if other_obj.id not in [x.id for x in self.update_objs]:  # type: ignore
                    self.update_objs.append(other_obj)

            elif diff.status == "NEW":
                if my_obj is None:
                    # keep IDs comparison here, otherwise it will break with actionobjects
                    if other_obj.id not in [x.id for x in self.create_objs]:  # type: ignore
                        self.create_objs.append(other_obj)

                elif other_obj is None:
                    # keep IDs comparison here, otherwise it will break with actionobjects
                    if my_obj.id not in [x.id for x in self.delete_objs]:
                        self.delete_objs.append(my_obj)

        if self.alias == "low":
            self.new_permissions.extend(sync_decision.new_permissions_lowside)
            self.new_storage_permissions.extend(
                sync_decision.new_storage_permissions_lowside
            )
        elif self.alias == "high":
            self.new_storage_permissions.extend(
                sync_decision.new_storage_permissions_highside
            )
        else:
            raise ValueError("Invalid alias")

    def __repr__(self) -> str:
        return (
            f"ResolvedSyncState(\n"
            f"  create_objs={self.create_objs},\n"
            f"  update_objs={self.update_objs},\n"
            f"  delete_objs={self.delete_objs}\n"
            f"  new_permissions={self.new_permissions}\n"
            f")"
        )


def display_diff_object(obj_state: str | None) -> Panel:
    if obj_state is None:
        return Panel(Markdown("None"), box=box.ROUNDED, expand=False)
    return Panel(
        Markdown(f"```python\n{obj_state}\n```", code_theme="default"),
        box=box.ROUNDED,
        expand=False,
    )


def display_diff_hierarchy(diff_hierarchy: list[tuple[ObjectDiff, int]]) -> None:
    console = Console()

    for diff, level in diff_hierarchy:
        title = f"{diff.obj_type.__name__}({diff.object_id}) - State: {diff.status}"

        low_side_panel = display_diff_object(diff.low_state if diff.low_obj else None)
        low_side_panel.title = "Low side"
        low_side_panel.title_align = "left"
        high_side_panel = display_diff_object(
            diff.high_state if diff.high_obj is not None else None
        )
        high_side_panel.title = "High side"
        high_side_panel.title_align = "left"

        grouped_panels = Group(low_side_panel, high_side_panel)

        diff_panel = Panel(
            grouped_panels,
            title=title,
            title_align="left",
            box=box.ROUNDED,
            expand=False,
            padding=(1, 2),
        )

        if level > 0:
            diff_panel = Padding(diff_panel, (0, 0, 0, 5 * level))

        console.print(diff_panel)



# future
from __future__ import annotations

# stdlib
from pathlib import Path
import re
from typing import TYPE_CHECKING
from typing import cast

# third party
from hagrid.orchestra import NodeHandle
from loguru import logger
from tqdm import tqdm

# relative
from ..abstract_node import NodeSideType
from ..img.base64 import base64read
from ..serde.serializable import serializable
from ..service.action.action_object import ActionObject
from ..service.code_history.code_history import CodeHistoriesDict
from ..service.code_history.code_history import UsersCodeHistoriesDict
from ..service.dataset.dataset import Contributor
from ..service.dataset.dataset import CreateAsset
from ..service.dataset.dataset import CreateDataset
from ..service.response import SyftError
from ..service.response import SyftSuccess
from ..service.sync.diff_state import ResolvedSyncState
from ..service.sync.sync_state import SyncState
from ..service.user.roles import Roles
from ..service.user.user import UserView
from ..service.user.user_roles import ServiceRole
from ..types.blob_storage import BlobFile
from ..types.uid import UID
from ..util.fonts import fonts_css
from ..util.util import get_mb_size
from ..util.util import prompt_warning_message
from .api import APIModule
from .client import SyftClient
from .client import login
from .client import login_as_guest
from .connection import NodeConnection
from .protocol import SyftProtocol

if TYPE_CHECKING:
    # relative
    from ..service.project.project import Project


def _get_files_from_glob(glob_path: str) -> list[Path]:
    files = Path().glob(glob_path)
    return [f for f in files if f.is_file() and not f.name.startswith(".")]


def _get_files_from_dir(dir: Path, recursive: bool) -> list:
    files = dir.rglob("*") if recursive else dir.iterdir()
    return [f for f in files if not f.name.startswith(".") and f.is_file()]


def _contains_subdir(dir: Path) -> bool:
    for item in dir.iterdir():
        if item.is_dir():
            return True
    return False


def add_default_uploader(
    user: UserView, obj: CreateDataset | CreateAsset
) -> CreateDataset | CreateAsset:
    uploader = None
    for contributor in obj.contributors:
        if contributor.role == str(Roles.UPLOADER):
            uploader = contributor
            break

    if uploader is None:
        uploader = Contributor(
            role=str(Roles.UPLOADER),
            name=user.name,
            email=user.email,
        )
        obj.contributors.add(uploader)

    obj.uploader = uploader
    return obj


@serializable()
class DomainClient(SyftClient):
    def __repr__(self) -> str:
        return f"<DomainClient: {self.name}>"

    def upload_dataset(self, dataset: CreateDataset) -> SyftSuccess | SyftError:
        # relative
        from ..types.twin_object import TwinObject

        if self.users is None:
            return SyftError(f"can't get user service for {self}")

        user = self.users.get_current_user()
        dataset = add_default_uploader(user, dataset)
        for i in range(len(dataset.asset_list)):
            asset = dataset.asset_list[i]
            dataset.asset_list[i] = add_default_uploader(user, asset)

        dataset._check_asset_must_contain_mock()
        dataset_size: float = 0.0

        # TODO: Refactor so that object can also be passed to generate warnings

        self.api.connection = cast(NodeConnection, self.api.connection)

        metadata = self.api.connection.get_node_metadata(self.api.signing_key)

        if (
            metadata.show_warnings
            and metadata.node_side_type == NodeSideType.HIGH_SIDE.value
        ):
            message = (
                "You're approving a request on "
                f"{metadata.node_side_type} side {metadata.node_type} "
                "which may host datasets with private information."
            )
            prompt_warning_message(message=message, confirm=True)

        for asset in tqdm(dataset.asset_list):
            print(f"Uploading: {asset.name}")
            try:
                twin = TwinObject(
                    private_obj=asset.data,
                    mock_obj=asset.mock,
                    syft_node_location=self.id,
                    syft_client_verify_key=self.verify_key,
                )
                twin._save_to_blob_storage()
            except Exception as e:
                return SyftError(message=f"Failed to create twin. {e}")
            response = self.api.services.action.set(twin)
            if isinstance(response, SyftError):
                print(f"Failed to upload asset\n: {asset}")
                return response
            asset.action_id = twin.id
            asset.node_uid = self.id
            dataset_size += get_mb_size(asset.data)
        dataset.mb_size = dataset_size
        valid = dataset.check()
        if isinstance(valid, SyftError):
            return valid
        return self.api.services.dataset.add(dataset=dataset)

    # def get_permissions_for_other_node(
    #     self,
    #     items: list[Union[ActionObject, SyftObject]],
    # ) -> dict:
    #     if len(items) > 0:
    #         if not len({i.syft_node_location for i in items}) == 1 or (
    #             not len({i.syft_client_verify_key for i in items}) == 1
    #         ):
    #             raise ValueError("permissions from different nodes")
    #         item = items[0]
    #         api = APIRegistry.api_for(
    #             item.syft_node_location, item.syft_client_verify_key
    #         )
    #         if api is None:
    #             raise ValueError(
    #                 f"Can't access the api. Please log in to {item.syft_node_location}"
    #             )
    #         return api.services.sync.get_permissions(items)
    #     else:
    #         return {}

    def get_sync_state(self) -> SyncState | SyftError:
        state: SyncState = self.api.services.sync._get_state()
        for uid, obj in state.objects.items():
            if isinstance(obj, ActionObject):
                state.objects[uid] = obj.refresh_object()
        return state

    def apply_state(self, resolved_state: ResolvedSyncState) -> SyftSuccess | SyftError:
        if len(resolved_state.delete_objs):
            raise NotImplementedError("TODO implement delete")
        items = resolved_state.create_objs + resolved_state.update_objs

        action_objects = [x for x in items if isinstance(x, ActionObject)]
        # permissions = self.get_permissions_for_other_node(items)

        permissions: dict[UID, set[str]] = {}
        for p in resolved_state.new_permissions:
            if p.uid in permissions:
                permissions[p.uid].add(p.permission_string)
            else:
                permissions[p.uid] = {p.permission_string}

        storage_permissions: dict[UID, set[UID]] = {}
        for sp in resolved_state.new_storage_permissions:
            if sp.uid in storage_permissions:
                storage_permissions[sp.uid].add(sp.node_uid)
            else:
                storage_permissions[sp.uid] = {sp.node_uid}

        for action_object in action_objects:
            # NOTE permissions are added separately server side
            action_object._send(self, add_storage_permission=False)

        res = self.api.services.sync.sync_items(
            items,
            permissions,
            storage_permissions,
        )
        if isinstance(res, SyftError):
            return res

        # Add updated node state to store to have a previous_state for next sync
        new_state = self.api.services.sync._get_state(add_to_store=True)
        if isinstance(new_state, SyftError):
            return new_state

        self._fetch_api(self.credentials)
        return res

    def upload_files(
        self,
        file_list: BlobFile | list[BlobFile] | str | list[str] | Path | list[Path],
        allow_recursive: bool = False,
        show_files: bool = False,
    ) -> SyftSuccess | SyftError:
        if not file_list:
            return SyftError(message="No files to upload")

        if not isinstance(file_list, list):
            file_list = [file_list]  # type: ignore[assignment]
        file_list = cast(list, file_list)

        expanded_file_list: list[BlobFile | Path] = []

        for file in file_list:
            if isinstance(file, BlobFile):
                expanded_file_list.append(file)
                continue

            path = Path(file)

            if re.search(r"[\*\?\[]", str(path)):
                expanded_file_list.extend(_get_files_from_glob(str(path)))
            elif path.is_dir():
                if not allow_recursive and _contains_subdir(path):
                    res = input(
                        f"Do you want to include all files recursively in {path.absolute()}? [y/n]: "
                    ).lower()
                    print(
                        f'{"Recursively uploading all files" if res == "y" else "Uploading files"} in {path.absolute()}'
                    )
                    allow_recursive = res == "y"
                expanded_file_list.extend(_get_files_from_dir(path, allow_recursive))
            elif path.exists():
                expanded_file_list.append(path)

        if not expanded_file_list:
            return SyftError(message="No files to upload were found")

        print(
            f"Uploading {len(expanded_file_list)} {'file' if len(expanded_file_list) == 1 else 'files'}:"
        )

        if show_files:
            for file in expanded_file_list:
                if isinstance(file, BlobFile):
                    print(file.path or file.file_name)
                else:
                    print(file.absolute())

        try:
            result = []
            for file in expanded_file_list:
                if not isinstance(file, BlobFile):
                    file = BlobFile(path=file, file_name=file.name)
                print("Uploading", file.file_name)
                if not file.uploaded:
                    file.upload_to_blobstorage(self)
                result.append(file)

            return ActionObject.from_obj(result).send(self)
        except Exception as err:
            logger.debug("upload_files: Error creating action_object: {}", err)
            return SyftError(message=f"Failed to upload files: {err}")

    def connect_to_gateway(
        self,
        via_client: SyftClient | None = None,
        url: str | None = None,
        port: int | None = None,
        handle: NodeHandle | None = None,  # noqa: F821
        email: str | None = None,
        password: str | None = None,
        protocol: str | SyftProtocol = SyftProtocol.HTTP,
    ) -> SyftSuccess | SyftError | None:
        if isinstance(protocol, str):
            protocol = SyftProtocol(protocol)

        if via_client is not None:
            client = via_client
        elif handle is not None:
            client = handle.client
        else:
            client = (
                login_as_guest(url=url, port=port)
                if email is None
                else login(url=url, port=port, email=email, password=password)
            )
            if isinstance(client, SyftError):
                return client

        res = self.exchange_route(client, protocol=protocol)
        if isinstance(res, SyftSuccess):
            if self.metadata:
                return SyftSuccess(
                    message=f"Connected {self.metadata.node_type} to {client.name} gateway"
                )
            else:
                return SyftSuccess(message=f"Connected to {client.name} gateway")
        return res

    @property
    def data_subject_registry(self) -> APIModule | None:
        if self.api.has_service("data_subject"):
            return self.api.services.data_subject
        return None

    @property
    def code(self) -> APIModule | None:
        # if self.api.refresh_api_callback is not None:
        #     self.api.refresh_api_callback()
        if self.api.has_service("code"):
            return self.api.services.code
        return None

    @property
    def worker(self) -> APIModule | None:
        if self.api.has_service("worker"):
            return self.api.services.worker
        return None

    @property
    def requests(self) -> APIModule | None:
        if self.api.has_service("request"):
            return self.api.services.request
        return None

    @property
    def datasets(self) -> APIModule | None:
        if self.api.has_service("dataset"):
            return self.api.services.dataset
        return None

    @property
    def projects(self) -> APIModule | None:
        if self.api.has_service("project"):
            return self.api.services.project
        return None

    @property
    def code_history_service(self) -> APIModule | None:
        if self.api is not None and self.api.has_service("code_history"):
            return self.api.services.code_history
        return None

    @property
    def code_history(self) -> CodeHistoriesDict:
        return self.api.services.code_history.get_history()

    @property
    def code_histories(self) -> UsersCodeHistoriesDict:
        return self.api.services.code_history.get_histories()

    @property
    def images(self) -> APIModule | None:
        if self.api.has_service("worker_image"):
            return self.api.services.worker_image
        return None

    @property
    def worker_pools(self) -> APIModule | None:
        if self.api.has_service("worker_pool"):
            return self.api.services.worker_pool
        return None

    @property
    def worker_images(self) -> APIModule | None:
        if self.api.has_service("worker_image"):
            return self.api.services.worker_image
        return None

    @property
    def sync(self) -> APIModule | None:
        if self.api.has_service("sync"):
            return self.api.services.sync
        return None

    @property
    def code_status(self) -> APIModule | None:
        if self.api.has_service("code_status"):
            return self.api.services.code_status
        return None

    @property
    def output(self) -> APIModule | None:
        if self.api.has_service("output"):
            return self.api.services.output
        return None

    def get_project(
        self,
        name: str | None = None,
        uid: UID | None = None,
    ) -> Project | None:
        """Get project by name or UID"""

        if not self.api.has_service("project"):
            return None

        if name:
            return self.api.services.project.get_by_name(name)

        elif uid:
            return self.api.services.project.get_by_uid(uid)

        return self.api.services.project.get_all()

    def _repr_html_(self) -> str:
        guest_commands = """
        <li><span class='syft-code-block'>&lt;your_client&gt;.datasets</span> - list datasets</li>
        <li><span class='syft-code-block'>&lt;your_client&gt;.code</span> - list code</li>
        <li><span class='syft-code-block'>&lt;your_client&gt;.login</span> - list projects</li>
        <li>
            <span class='syft-code-block'>&lt;your_client&gt;.code.submit?</span> - display function signature
        </li>"""
        ds_commands = """
        <li><span class='syft-code-block'>&lt;your_client&gt;.datasets</span> - list datasets</li>
        <li><span class='syft-code-block'>&lt;your_client&gt;.code</span> - list code</li>
        <li><span class='syft-code-block'>&lt;your_client&gt;.projects</span> - list projects</li>
        <li>
            <span class='syft-code-block'>&lt;your_client&gt;.code.submit?</span> - display function signature
        </li>"""

        do_commands = """
        <li><span class='syft-code-block'>&lt;your_client&gt;.projects</span> - list projects</li>
        <li><span class='syft-code-block'>&lt;your_client&gt;.requests</span> - list requests</li>
        <li><span class='syft-code-block'>&lt;your_client&gt;.users</span> - list users</li>
        <li>
            <span class='syft-code-block'>&lt;your_client&gt;.requests.submit?</span> - display function signature
        </li>"""

        # TODO: how to select ds/do commands based on self.__user_role

        if (
            self.user_role.value == ServiceRole.NONE.value
            or self.user_role.value == ServiceRole.GUEST.value
        ):
            commands = guest_commands
        elif (
            self.user_role is not None
            and self.user_role.value == ServiceRole.DATA_SCIENTIST.value
        ):
            commands = ds_commands
        elif (
            self.user_role is not None
            and self.user_role.value >= ServiceRole.DATA_OWNER.value
        ):
            commands = do_commands

        command_list = f"""
        <ul style='padding-left: 1em;'>
            {commands}
        </ul>
        """

        small_grid_symbol_logo = base64read("small-grid-symbol-logo.png")

        url = getattr(self.connection, "url", None)
        node_details = f"<strong>URL:</strong> {url}<br />" if url else ""
        if self.metadata is not None:
            node_details += f"<strong>Node Type:</strong> {self.metadata.node_type.capitalize()}<br />"
            node_side_type = (
                "Low Side"
                if self.metadata.node_side_type == NodeSideType.LOW_SIDE.value
                else "High Side"
            )
            node_details += f"<strong>Node Side Type:</strong> {node_side_type}<br />"
            node_details += (
                f"<strong>Syft Version:</strong> {self.metadata.syft_version}<br />"
            )

        return f"""
        <style>
            {fonts_css}

            .syft-container {{
                padding: 5px;
                font-family: 'Open Sans';
            }}
            .syft-alert-info {{
                color: #1F567A;
                background-color: #C2DEF0;
                border-radius: 4px;
                padding: 5px;
                padding: 13px 10px
            }}
            .syft-code-block {{
                background-color: #f7f7f7;
                border: 1px solid #cfcfcf;
                padding: 0px 2px;
            }}
            .syft-space {{
                margin-top: 1em;
            }}
        </style>
        <div class="syft-client syft-container">
            <img src="{small_grid_symbol_logo}" alt="Logo"
            style="width:48px;height:48px;padding:3px;">
            <h2>Welcome to {self.name}</h2>
            <div class="syft-space">
                {node_details}
            </div>
            <div class='syft-alert-info syft-space'>
                &#9432;&nbsp;
                This domain is run by the library PySyft to learn more about how it works visit
                <a href="https://github.com/OpenMined/PySyft">github.com/OpenMined/PySyft</a>.
            </div>
            <h4>Commands to Get Started</h4>
            {command_list}
        </div><br />
        """



# stdlib
from time import sleep

# relative
from ..service.action.action_permissions import ActionObjectPermission
from ..service.action.action_permissions import ActionPermission
from ..service.action.action_permissions import StoragePermission
from ..service.code.user_code import UserCode
from ..service.job.job_stash import Job
from ..service.sync.diff_state import NodeDiff
from ..service.sync.diff_state import ObjectDiff
from ..service.sync.diff_state import ObjectDiffBatch
from ..service.sync.diff_state import ResolvedSyncState
from ..service.sync.diff_state import SyncDecision
from ..service.sync.sync_state import SyncState


def compare_states(low_state: SyncState, high_state: SyncState) -> NodeDiff:
    return NodeDiff.from_sync_state(low_state=low_state, high_state=high_state)


def get_user_input_for_resolve() -> str | None:
    print(
        "Do you want to keep the low state or the high state for these objects? choose 'low' or 'high'"
    )

    while True:
        decision = input()
        decision = decision.lower()

        if decision in ["low", "high"]:
            return decision
        else:
            print("Please choose between `low` or `high`")


def resolve(
    state: NodeDiff,
    decision: str | None = None,
    share_private_objects: bool = False,
    ask_for_input: bool = True,
) -> tuple[ResolvedSyncState, ResolvedSyncState]:
    # TODO: only add permissions for objects where we manually give permission
    # Maybe default read permission for some objects (high -> low)
    resolved_state_low = ResolvedSyncState(node_uid=state.low_node_uid, alias="low")
    resolved_state_high = ResolvedSyncState(node_uid=state.high_node_uid, alias="high")

    for batch_diff in state.hierarchies:
        batch_decision = decision
        if all(diff.status == "SAME" for diff in batch_diff.diffs):
            # Hierarchy has no diffs
            continue

        print(batch_diff.__repr__())

        if batch_decision is None:
            batch_decision = get_user_input_for_resolve()

        sync_decisions: list[SyncDecision] = get_sync_decisions_for_batch_items(
            batch_diff,
            batch_decision,
            share_private_objects=share_private_objects,
            ask_for_input=ask_for_input,
        )

        print(f"Decision: Syncing {len(batch_diff)} objects from {batch_decision} side")

        for sync_decision in sync_decisions:
            resolved_state_low.add_sync_decision(sync_decision)
            resolved_state_high.add_sync_decision(sync_decision)

        print()
        print("=" * 100)
        print()

    return resolved_state_low, resolved_state_high


def get_sync_decisions_for_batch_items(
    batch_diff: ObjectDiffBatch,
    decision: str,
    share_private_objects: bool = False,
    ask_for_input: bool = True,
) -> list[SyncDecision]:
    sync_decisions: list[SyncDecision] = []

    unpublished_private_high_diffs: list[ObjectDiff] = []
    for diff in batch_diff.diffs:
        is_high_private_object = (
            diff.high_obj is not None and diff.high_obj._has_private_sync_attrs()
        )
        is_low_published_object = diff.low_node_uid in diff.low_storage_permissions
        if is_high_private_object and not is_low_published_object:
            unpublished_private_high_diffs.append(diff)

    user_codes_high: list[UserCode] = [
        diff.high_obj
        for diff in batch_diff.diffs
        if isinstance(diff.high_obj, UserCode)
    ]
    if len(user_codes_high) > 1:
        raise ValueError("too many user codes")
    if len(user_codes_high) == 0:
        user_code_high = None
    else:
        user_code_high = user_codes_high[0]

    if user_code_high is None and len(unpublished_private_high_diffs):
        raise ValueError("Found unpublished private objects without user code")

    if share_private_objects:
        private_high_diffs_to_share = unpublished_private_high_diffs
    elif ask_for_input:
        private_high_diffs_to_share = ask_user_input_permission(
            user_code_high, unpublished_private_high_diffs
        )
    else:
        private_high_diffs_to_share = []

    for diff in batch_diff.diffs:
        is_unpublished_private_diff = diff in unpublished_private_high_diffs
        has_share_decision = diff in private_high_diffs_to_share

        if isinstance(diff.high_obj, Job):
            if user_code_high is None:
                raise ValueError("Job without user code")
            # Jobs are always shared
            new_permissions_low_side = [
                ActionObjectPermission(
                    uid=diff.object_id,
                    permission=ActionPermission.READ,
                    credentials=user_code_high.user_verify_key,
                )
            ]
            mockify = False

        elif is_unpublished_private_diff and has_share_decision:
            # private + want to share
            new_permissions_low_side = [
                ActionObjectPermission(
                    uid=diff.object_id,
                    permission=ActionPermission.READ,
                    credentials=user_code_high.user_verify_key,  # type: ignore
                )
            ]
            mockify = False

        elif is_unpublished_private_diff and not has_share_decision:
            # private + do not share
            new_permissions_low_side = []
            mockify = True

        else:
            # any other object is shared
            new_permissions_low_side = []
            mockify = False

        new_storage_permissions_lowside = []
        if not mockify:
            new_storage_permissions_lowside = [
                StoragePermission(uid=diff.object_id, node_uid=diff.low_node_uid)
            ]

        # Always share to high_side
        if diff.status == "NEW" and diff.high_obj is None:
            new_storage_permissions_highside = [
                StoragePermission(uid=diff.object_id, node_uid=diff.high_node_uid)
            ]
        else:
            new_storage_permissions_highside = []

        sync_decisions.append(
            SyncDecision(
                diff=diff,
                decision=decision,
                new_permissions_lowside=new_permissions_low_side,
                new_storage_permissions_lowside=new_storage_permissions_lowside,
                new_storage_permissions_highside=new_storage_permissions_highside,
                mockify=mockify,
            )
        )

    return sync_decisions


QUESTION_SHARE_PRIVATE_OBJS = """You currently have the following private objects:

{objects_str}

Do you want to share some of these private objects? If so type the first 3 characters of the id e.g. 'abc'.
If you dont want to share any more private objects, type "no"
"""

CONFIRMATION_SHARE_PRIVATE_OBJ = """Setting permissions for {object_type} #{object_id} to share with {user_verify_key},
this will become effective when you call client.apply_state(<resolved_state>))
"""


def ask_user_input_permission(
    user_code: UserCode, all_private_high_diffs: list[ObjectDiff]
) -> list[ObjectDiff]:
    if len(all_private_high_diffs) == 0:
        return []

    user_verify_key = user_code.user_verify_key
    private_high_diffs_to_share = []
    print(
        f"""This batch of updates contains new private objects on the high side that you may want \
    to share with user {user_verify_key}."""
    )

    remaining_private_high_diffs = all_private_high_diffs[:]
    while len(remaining_private_high_diffs):
        objects_str = "\n".join(
            [
                f"{diff.object_type} #{diff.object_id}"
                for diff in remaining_private_high_diffs
            ]
        )
        print(QUESTION_SHARE_PRIVATE_OBJS.format(objects_str=objects_str), flush=True)

        sleep(0.1)
        res = input()
        if res == "no":
            break
        elif len(res) >= 3:
            matches = [
                diff
                for diff in remaining_private_high_diffs
                if str(diff.object_id).startswith(res)
            ]
            if len(matches) == 0:
                print("Invalid input")
                continue
            elif len(matches) == 1:
                diff = matches[0]
                print()
                print("=" * 100)
                print()
                print(
                    CONFIRMATION_SHARE_PRIVATE_OBJ.format(
                        object_type=diff.object_type,
                        object_id=diff.object_id,
                        user_verify_key=user_verify_key,
                    )
                )

                remaining_private_high_diffs.remove(diff)
                private_high_diffs_to_share.append(diff)

            else:
                print("Found multiple matches for provided id, exiting")
                break
        else:
            print("invalid input")

    return private_high_diffs_to_share



# stdlib

# third party
from result import Result

# relative
from ...node.credentials import SyftVerifyKey
from ...serde.serializable import serializable
from ...store.document_store import BaseUIDStoreStash
from ...store.document_store import DocumentStore
from ...store.document_store import PartitionSettings
from ...store.document_store import QueryKeys
from ...store.document_store import UIDPartitionKey
from ...types.uid import UID
from ...util.telemetry import instrument
from ..context import AuthedServiceContext
from ..response import SyftError
from ..service import AbstractService
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user_roles import ADMIN_ROLE_LEVEL
from ..user.user_roles import GUEST_ROLE_LEVEL
from .user_code import UserCodeStatusCollection


@instrument
@serializable()
class StatusStash(BaseUIDStoreStash):
    object_type = UserCodeStatusCollection
    settings: PartitionSettings = PartitionSettings(
        name=UserCodeStatusCollection.__canonical_name__,
        object_type=UserCodeStatusCollection,
    )

    def __init__(self, store: DocumentStore) -> None:
        super().__init__(store)
        self.store = store
        self.settings = self.settings
        self._object_type = self.object_type

    def get_by_uid(
        self, credentials: SyftVerifyKey, uid: UID
    ) -> Result[UserCodeStatusCollection, str]:
        qks = QueryKeys(qks=[UIDPartitionKey.with_obj(uid)])
        return self.query_one(credentials=credentials, qks=qks)


@instrument
@serializable()
class UserCodeStatusService(AbstractService):
    store: DocumentStore
    stash: StatusStash

    def __init__(self, store: DocumentStore):
        self.store = store
        self.stash = StatusStash(store=store)

    @service_method(path="code_status.create", name="create", roles=ADMIN_ROLE_LEVEL)
    def create(
        self,
        context: AuthedServiceContext,
        status: UserCodeStatusCollection,
    ) -> UserCodeStatusCollection | SyftError:
        result = self.stash.set(
            credentials=context.credentials,
            obj=status,
        )
        if result.is_ok():
            return result.ok()
        return SyftError(message=result.err())

    @service_method(
        path="code_status.get_by_uid", name="get_by_uid", roles=GUEST_ROLE_LEVEL
    )
    def get_status(
        self, context: AuthedServiceContext, uid: UID
    ) -> UserCodeStatusCollection | SyftError:
        """Get the status of a user code item"""
        result = self.stash.get_by_uid(context.credentials, uid=uid)
        if result.is_ok():
            return result.ok()
        return SyftError(message=result.err())

    @service_method(path="code_status.get_all", name="get_all", roles=ADMIN_ROLE_LEVEL)
    def get_all(
        self, context: AuthedServiceContext
    ) -> list[UserCodeStatusCollection] | SyftError:
        """Get all user code item statuses"""
        result = self.stash.get_all(context.credentials)
        if result.is_ok():
            return result.ok()
        return SyftError(message=result.err())


TYPE_TO_SERVICE[UserCodeStatusCollection] = UserCodeStatusService



VEILID_SERVICE_URL = "http://veilid:80"
# Service name of our traefik service
# TODO: Remove this once when we remove reverse proxy in Veilid Connection
VEILID_SYFT_PROXY_URL = "http://proxy:80"
HEALTHCHECK_ENDPOINT = "/healthcheck"
GEN_VLD_KEY_ENDPOINT = "/generate_vld_key"
RET_VLD_KEY_ENDPOINT = "/retrieve_vld_key"
VEILID_PROXY_PATH = "/proxy"



# stdlib
from collections.abc import Callable
from collections.abc import Iterable
from enum import Enum
from functools import partial
import os
from pathlib import Path
import tempfile
from typing import Any

# third party
import matplotlib.pyplot as plt
import networkx as nx
from pydantic import Field
from pydantic import field_validator
from result import Err
from result import Ok
from result import Result
from typing_extensions import Self

# relative
from ...node.credentials import SyftVerifyKey
from ...serde.deserialize import _deserialize
from ...serde.serializable import serializable
from ...serde.serialize import _serialize
from ...store.document_store import QueryKey
from ...store.document_store import QueryKeys
from ...store.document_store import StoreClientConfig
from ...store.document_store import StoreConfig
from ...store.locks import LockingConfig
from ...store.locks import SyftLock
from ...store.locks import ThreadingLockingConfig
from ...types.datetime import DateTime
from ...types.syft_object import PartialSyftObject
from ...types.syft_object import SYFT_OBJECT_VERSION_2
from ...types.syft_object import SyftObject
from ...types.uid import UID
from .action_object import Action
from .action_object import ActionObject


@serializable()
class ExecutionStatus(Enum):
    PROCESSING = 0
    DONE = 1
    FAILED = 2


@serializable()
class NodeType(Enum):
    ACTION = Action
    ACTION_OBJECT = ActionObject


@serializable()
class NodeActionData(SyftObject):
    __canonical_name__ = "NodeActionData"
    __version__ = SYFT_OBJECT_VERSION_2

    id: UID | None = None  # type: ignore[assignment]
    type: NodeType
    status: ExecutionStatus = ExecutionStatus.PROCESSING
    retry: int = 0
    created_at: DateTime = Field(default_factory=DateTime.now)
    updated_at: DateTime = Field(default_factory=DateTime.now)
    user_verify_key: SyftVerifyKey
    is_mutated: bool = False  # denotes that this node has been mutated
    is_mutagen: bool = False  # denotes that this node is causing a mutation
    next_mutagen_node: UID | None = None  # next neighboring mutagen node
    last_nm_mutagen_node: UID | None = None  # last non mutated mutagen node

    @classmethod
    def from_action(cls, action: Action, credentials: SyftVerifyKey) -> Self:
        is_mutagen = action.remote_self is not None and (
            action.remote_self == action.result_id
        )
        return cls(
            id=action.id,
            type=NodeType.ACTION,
            user_verify_key=credentials,
            is_mutagen=is_mutagen,
        )

    @classmethod
    def from_action_obj(
        cls, action_obj: ActionObject, credentials: SyftVerifyKey
    ) -> Self:
        return cls(
            id=action_obj.id,
            type=NodeType.ACTION_OBJECT,
            user_verify_key=credentials,
        )

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, NodeActionData):
            raise NotImplementedError(
                "Comparisions can be made with NodeActionData type objects only."
            )
        return hash(self) == hash(other)

    def __repr__(self) -> str:
        return self._repr_debug_()


@serializable()
class NodeActionDataUpdate(PartialSyftObject):
    __canonical_name__ = "NodeActionDataUpdate"
    __version__ = SYFT_OBJECT_VERSION_2

    id: UID
    type: NodeType
    status: ExecutionStatus
    retry: int
    created_at: DateTime
    updated_at: DateTime = Field(default_factory=DateTime.now)
    credentials: SyftVerifyKey
    is_mutated: bool
    is_mutagen: bool
    next_mutagen_node: UID  # next neighboring mutagen node
    last_nm_mutagen_node: UID  # last non mutated mutagen node


@serializable()
class BaseGraphStore:
    graph_type: Any
    client_config: StoreClientConfig | None

    def set(self, uid: Any, data: Any) -> None:
        raise NotImplementedError

    def get(self, uid: Any) -> Any:
        raise NotImplementedError

    def delete(self, uid: Any) -> None:
        raise NotImplementedError

    def find_neighbors(self, uid: Any) -> list | None:
        raise NotImplementedError

    def update(self, uid: Any, data: Any) -> None:
        raise NotImplementedError

    def add_edge(self, parent: Any, child: Any) -> None:
        raise NotImplementedError

    def remove_edge(self, parent: Any, child: Any) -> None:
        raise NotImplementedError

    def nodes(self) -> Any:
        raise NotImplementedError

    def edges(self) -> Any:
        raise NotImplementedError

    def visualize(self, seed: int, figsize: tuple) -> None:
        raise NotImplementedError

    def save(self) -> None:
        raise NotImplementedError

    def get_predecessors(self, uid: UID) -> list:
        raise NotImplementedError

    def get_successors(self, uid: UID) -> list:
        raise NotImplementedError

    def exists(self, uid: Any) -> bool:
        raise NotImplementedError

    def subgraph(self, qks: QueryKeys) -> Any:
        raise NotImplementedError

    def topological_sort(self, subgraph: Any) -> Any:
        raise NotImplementedError


@serializable()
class InMemoryStoreClientConfig(StoreClientConfig):
    filename: str = "action_graph.bytes"
    path: str | Path = Field(default_factory=tempfile.gettempdir)

    # We need this in addition to Field(default_factory=...)
    # so users can still do InMemoryStoreClientConfig(path=None)
    @field_validator("path", mode="before")
    @classmethod
    def __default_path(cls, path: str | Path | None) -> str | Path:
        if path is None:
            return tempfile.gettempdir()
        return path

    @property
    def file_path(self) -> Path:
        return Path(self.path) / self.filename


@serializable(without=["_lock"])
class NetworkXBackingStore(BaseGraphStore):
    def __init__(self, store_config: StoreConfig, reset: bool = False) -> None:
        if store_config.client_config:
            self.path_str = store_config.client_config.file_path.as_posix()
        else:
            self.path_str = ""
        if not reset and os.path.exists(self.path_str):
            self._db = self._load_from_path(self.path_str)
        else:
            self._db = nx.DiGraph()

        self.locking_config = store_config.locking_config
        self._lock: SyftLock | None = None

    @property
    def lock(self) -> SyftLock:
        if not hasattr(self, "_lock") or self._lock is None:
            self._lock = SyftLock(self.locking_config)
        return self._lock

    @property
    def db(self) -> nx.Graph:
        return self._db

    def _thread_safe_cbk(
        self, cbk: Callable, *args: Any, **kwargs: Any
    ) -> Result[Any, str]:
        # TODO copied method from document_store, have it in one place and reuse?
        locked = self.lock.acquire(blocking=True)
        if not locked:
            return Err(
                f"Failed to acquire lock for the operation {self.lock.lock_name} ({self.lock._lock})"
            )
        try:
            result = cbk(*args, **kwargs)
        except BaseException as e:
            result = Err(str(e))
        self.lock.release()

        return result

    def set(self, uid: UID, data: Any) -> None:
        self._thread_safe_cbk(self._set, uid=uid, data=data)

    def _set(self, uid: UID, data: Any) -> None:
        if self.exists(uid=uid):
            self.update(uid=uid, data=data)
        else:
            self.db.add_node(uid, data=data)
        self.save()

    def get(self, uid: UID) -> Any:
        node_data = self.db.nodes.get(uid)
        return node_data.get("data")

    def exists(self, uid: Any) -> bool:
        return uid in self.nodes()

    def delete(self, uid: UID) -> None:
        self._thread_safe_cbk(self._delete, uid=uid)

    def _delete(self, uid: UID) -> None:
        if self.exists(uid=uid):
            self.db.remove_node(uid)
        self.save()

    def find_neighbors(self, uid: UID) -> list | None:
        if self.exists(uid=uid):
            neighbors = self.db.neighbors(uid)
            return neighbors
        return None

    def update(self, uid: UID, data: Any) -> None:
        self._thread_safe_cbk(self._update, uid=uid, data=data)

    def _update(self, uid: UID, data: Any) -> None:
        if self.exists(uid=uid):
            self.db.nodes[uid]["data"] = data
        self.save()

    def add_edge(self, parent: Any, child: Any) -> None:
        self._thread_safe_cbk(self._add_edge, parent=parent, child=child)

    def _add_edge(self, parent: Any, child: Any) -> None:
        self.db.add_edge(parent, child)
        self.save()

    def remove_edge(self, parent: Any, child: Any) -> None:
        self._thread_safe_cbk(self._remove_edge, parent=parent, child=child)

    def _remove_edge(self, parent: Any, child: Any) -> None:
        self.db.remove_edge(parent, child)
        self.save()

    def visualize(self, seed: int = 3113794652, figsize: tuple = (20, 10)) -> None:
        plt.figure(figsize=figsize)
        pos = nx.spring_layout(self.db, seed=seed)
        return nx.draw_networkx(self.db, pos=pos, with_labels=True)

    def nodes(self) -> Iterable:
        return self.db.nodes(data=True)

    def edges(self) -> Iterable:
        return self.db.edges()

    def get_predecessors(self, uid: UID) -> list:
        return self.db.predecessors(uid)

    def get_successors(self, uid: UID) -> list:
        return self.db.successors(uid)

    def is_parent(self, parent: Any, child: Any) -> bool:
        parents = self.db.predecessors(child)
        return parent in parents

    def save(self) -> None:
        bytes = _serialize(self.db, to_bytes=True)
        with open(self.path_str, "wb") as f:
            f.write(bytes)

    def _filter_nodes_by(self, uid: UID, qks: QueryKeys) -> bool:
        node_data = self.db.nodes[uid]["data"]
        matches = []
        for qk in qks.all:
            matches.append(getattr(node_data, qk.key) == qk.value)
        # AND matches
        return all(matches)

    def subgraph(self, qks: QueryKeys) -> Any:
        filter_func = partial(self._filter_nodes_by, qks=qks)
        return nx.subgraph_view(self.db, filter_node=filter_func)

    def topological_sort(self, subgraph: Any) -> Any:
        return list(nx.topological_sort(subgraph))

    @staticmethod
    def _load_from_path(file_path: str) -> None:
        with open(file_path, "rb") as f:
            bytes = f.read()
        return _deserialize(blob=bytes, from_bytes=True)


@serializable()
class InMemoryGraphConfig(StoreConfig):
    __canonical_name__ = "InMemoryGraphConfig"

    store_type: type[BaseGraphStore] = NetworkXBackingStore
    client_config: StoreClientConfig = InMemoryStoreClientConfig()
    locking_config: LockingConfig = ThreadingLockingConfig()


@serializable()
class ActionGraphStore:
    pass


@serializable()
class InMemoryActionGraphStore(ActionGraphStore):
    __canonical_name__ = "InMemoryActionGraphStore"

    def __init__(self, store_config: StoreConfig, reset: bool = False):
        self.store_config: StoreConfig = store_config
        self.graph: BaseGraphStore = self.store_config.store_type(
            self.store_config, reset
        )

    def set(
        self,
        node: NodeActionData,
        credentials: SyftVerifyKey,
        parent_uids: list[UID] | None = None,
    ) -> Result[NodeActionData, str]:
        if self.graph.exists(uid=node.id):
            return Err(f"Node already exists in the graph: {node}")

        self.graph.set(uid=node.id, data=node)

        if parent_uids is None:
            parent_uids = []

        for parent_uid in parent_uids:
            result = self.add_edge(
                parent=parent_uid,
                child=node.id,
                credentials=credentials,
            )
            if result.is_err():
                return result

        return Ok(node)

    def get(
        self,
        uid: UID,
        credentials: SyftVerifyKey,
    ) -> Result[NodeActionData, str]:
        # 🟡 TODO: Add permission check
        if self.graph.exists(uid=uid):
            node_data = self.graph.get(uid=uid)
            return Ok(node_data)
        return Err(f"Node does not exists with id: {uid}")

    def delete(
        self,
        uid: UID,
        credentials: SyftVerifyKey,
    ) -> Result[bool, str]:
        # 🟡 TODO: Add permission checks
        if self.graph.exists(uid=uid):
            self.graph.delete(uid=uid)
            return Ok(True)
        return Err(f"Node does not exists with id: {uid}")

    def update(
        self,
        uid: UID,
        data: NodeActionDataUpdate,
        credentials: SyftVerifyKey,
    ) -> Result[NodeActionData, str]:
        # 🟡 TODO: Add permission checks
        node_data = self.graph.get(uid=uid)
        if node_data is not None:
            for key, val in data.to_dict(exclude_empty=True).items():
                setattr(node_data, key, val)
            self.graph.update(uid=uid, data=node_data)
            return Ok(node_data)
        return Err(f"Node does not exists for uid: {uid}")

    def update_non_mutated_successor(
        self,
        node_id: UID,
        nm_successor_id: UID,
        credentials: SyftVerifyKey,
    ) -> Result[NodeActionData, str]:
        """
        Used when a node is a mutagen and to update non-mutated
        successor for all nodes between node_id and nm_successor_id
        """
        node_data = self.graph.get(uid=node_id)

        data = NodeActionDataUpdate(
            next_mutagen_node=nm_successor_id,
            last_nm_mutagen_node=nm_successor_id,
            is_mutated=True,
        )

        if not node_data.is_mutated:
            # If current node is not mutated, then mark it as mutated
            return self.update(uid=node_id, data=data, credentials=credentials)
        else:
            # loop through successive mutagen nodes and
            # update their last_nm_mutagen_node id
            while node_id != nm_successor_id:
                node_data = self.graph.get(uid=node_id)

                # If node is the last added mutagen node,
                # then in that case its `next_mutagen_node` will be None
                # Therefore update its values to nm_successor_id
                next_mutagen_node = (
                    nm_successor_id
                    if node_data.next_mutagen_node is None
                    else node_data.next_mutagen_node
                )

                data = NodeActionDataUpdate(
                    last_nm_mutagen_node=nm_successor_id,
                    is_mutated=True,
                    next_mutagen_node=next_mutagen_node,
                )

                # Update each successive mutagen node
                result = self.update(
                    uid=node_id,
                    data=data,
                    credentials=credentials,
                )
                node_id = node_data.next_mutagen_node

            return result

    def _get_last_non_mutated_mutagen(
        self, credentials: SyftVerifyKey, uid: UID
    ) -> Result[UID, str]:
        node_data = self.graph.get(uid=uid)
        if node_data.is_mutated:
            return Ok(node_data.last_nm_mutagen_node)

        return Ok(uid)

    def add_edge(
        self,
        parent: UID,
        child: UID,
        credentials: SyftVerifyKey,
    ) -> Result[bool, str]:
        if not self.graph.exists(parent):
            return Err(f"Node does not exists for uid (parent): {parent}")

        if not self.graph.exists(child):
            return Err(f"Node does not exists for uid (child): {child}")

        result = self._get_last_non_mutated_mutagen(
            uid=parent,
            credentials=credentials,
        )

        if result.is_err():
            return result

        new_parent = result.ok()

        self.graph.add_edge(parent=new_parent, child=child)

        return Ok(True)

    def is_parent(self, parent: UID, child: UID) -> Result[bool, str]:
        if self.graph.exists(child):
            parents = self.graph.get_predecessors(child)
            result = parent in parents
            return Ok(result)
        return Err(f"Node doesn't exists for id: {child}")

    def query(
        self,
        qks: QueryKey | QueryKeys,
        credentials: SyftVerifyKey,
    ) -> Result[list[NodeActionData], str]:
        if isinstance(qks, QueryKey):
            qks = QueryKeys(qks=[qks])
        subgraph = self.graph.subgraph(qks=qks)
        return Ok(self.graph.topological_sort(subgraph=subgraph))

    def nodes(self, credentials: SyftVerifyKey) -> Result[list, str]:
        return Ok(self.graph.nodes())

    def edges(self, credentials: SyftVerifyKey) -> Result[list, str]:
        return Ok(self.graph.edges())



# stdlib
from collections.abc import Generator
from io import BytesIO
import math
from queue import Queue
import threading
from typing import Any

# third party
import boto3
from botocore.client import BaseClient as S3BaseClient
from botocore.client import ClientError as BotoClientError
from botocore.client import Config
import requests
from tqdm import tqdm
from typing_extensions import Self

# relative
from . import BlobDeposit
from . import BlobRetrieval
from . import BlobStorageClient
from . import BlobStorageClientConfig
from . import BlobStorageConfig
from . import BlobStorageConnection
from ...serde.serializable import serializable
from ...service.blob_storage.remote_profile import AzureRemoteProfile
from ...service.response import SyftError
from ...service.response import SyftSuccess
from ...service.service import from_api_or_context
from ...types.blob_storage import BlobStorageEntry
from ...types.blob_storage import CreateBlobStorageEntry
from ...types.blob_storage import SeaweedSecureFilePathLocation
from ...types.blob_storage import SecureFilePathLocation
from ...types.grid_url import GridURL
from ...types.syft_object import SYFT_OBJECT_VERSION_3
from ...util.constants import DEFAULT_TIMEOUT

WRITE_EXPIRATION_TIME = 900  # seconds
DEFAULT_FILE_PART_SIZE = (1024**3) * 5  # 5GB
DEFAULT_UPLOAD_CHUNK_SIZE = 819200


@serializable()
class SeaweedFSBlobDeposit(BlobDeposit):
    __canonical_name__ = "SeaweedFSBlobDeposit"
    __version__ = SYFT_OBJECT_VERSION_3

    urls: list[GridURL]
    size: int

    def write(self, data: BytesIO) -> SyftSuccess | SyftError:
        # relative
        from ...client.api import APIRegistry

        api = APIRegistry.api_for(
            node_uid=self.syft_node_location,
            user_verify_key=self.syft_client_verify_key,
        )

        etags = []

        try:
            no_lines = 0
            # this loops over the parts, we have multiple parts to allow for
            # concurrent uploads of a single file. (We are currently not using that)
            # a part may for instance be 5GB
            # parts are then splitted into chunks which are MBs (order of magnitude)
            part_size = math.ceil(self.size / len(self.urls))
            chunk_size = DEFAULT_UPLOAD_CHUNK_SIZE

            # this is the total nr of chunks in all parts
            total_iterations = math.ceil(part_size / chunk_size) * len(self.urls)

            with tqdm(
                total=total_iterations,
                desc=f"Uploading progress",  # noqa
            ) as pbar:
                for part_no, url in enumerate(
                    self.urls,
                    start=1,
                ):
                    if api is not None and api.connection is not None:
                        blob_url = api.connection.to_blob_route(
                            url.url_path, host=url.host_or_ip
                        )
                    else:
                        blob_url = url

                    # read a chunk untill we have read part_size
                    class PartGenerator:
                        def __init__(self) -> None:
                            self.no_lines = 0

                        def async_generator(
                            self, chunk_size: int = DEFAULT_UPLOAD_CHUNK_SIZE
                        ) -> Generator:
                            item_queue: Queue = Queue()
                            threading.Thread(
                                target=self.add_chunks_to_queue,
                                kwargs={"queue": item_queue, "chunk_size": chunk_size},
                                daemon=True,
                            ).start()
                            item = item_queue.get()
                            while item != 0:
                                yield item
                                pbar.update(1)
                                item = item_queue.get()

                        def add_chunks_to_queue(
                            self,
                            queue: Queue,
                            chunk_size: int = DEFAULT_UPLOAD_CHUNK_SIZE,
                        ) -> None:
                            """Creates a data geneator for the part"""
                            n = 0

                            while n * chunk_size <= part_size:
                                try:
                                    chunk = data.read(chunk_size)
                                    self.no_lines += chunk.count(b"\n")
                                    n += 1
                                    queue.put(chunk)
                                except BlockingIOError:
                                    # if end of file, stop
                                    queue.put(0)
                            # if end of part, stop
                            queue.put(0)

                    gen = PartGenerator()

                    response = requests.put(
                        url=str(blob_url),
                        data=gen.async_generator(chunk_size),
                        timeout=DEFAULT_TIMEOUT,
                        stream=True,
                    )

                    response.raise_for_status()
                    no_lines += gen.no_lines
                    etag = response.headers["ETag"]
                    etags.append({"ETag": etag, "PartNumber": part_no})

        except requests.RequestException as e:
            print(e)
            return SyftError(message=str(e))

        mark_write_complete_method = from_api_or_context(
            func_or_path="blob_storage.mark_write_complete",
            syft_node_location=self.syft_node_location,
            syft_client_verify_key=self.syft_client_verify_key,
        )
        if mark_write_complete_method is None:
            return SyftError(message="mark_write_complete_method is None")
        return mark_write_complete_method(
            etags=etags, uid=self.blob_storage_entry_id, no_lines=no_lines
        )


@serializable()
class SeaweedFSClientConfig(BlobStorageClientConfig):
    host: str
    port: int
    mount_port: int | None = None
    access_key: str
    secret_key: str
    region: str
    default_bucket_name: str = "defaultbucket"
    remote_profiles: dict[str, AzureRemoteProfile] = {}

    @property
    def endpoint_url(self) -> str:
        grid_url = GridURL(host_or_ip=self.host, port=self.port)
        return grid_url.url

    @property
    def mount_url(self) -> str:
        if self.mount_port is None:
            raise ValueError("Seaweed should be configured with a mount port to mount")
        return f"http://{self.host}:{self.mount_port}/configure_azure"


@serializable()
class SeaweedFSClient(BlobStorageClient):
    config: SeaweedFSClientConfig

    def connect(self) -> BlobStorageConnection:
        return SeaweedFSConnection(
            client=boto3.client(
                "s3",
                endpoint_url=self.config.endpoint_url,
                aws_access_key_id=self.config.access_key,
                aws_secret_access_key=self.config.secret_key,
                config=Config(signature_version="s3v4"),
                region_name=self.config.region,
            ),
            default_bucket_name=self.config.default_bucket_name,
            config=self.config,
        )


@serializable()
class SeaweedFSConnection(BlobStorageConnection):
    client: S3BaseClient
    default_bucket_name: str
    config: SeaweedFSClientConfig

    def __init__(
        self,
        client: S3BaseClient,
        default_bucket_name: str,
        config: SeaweedFSClientConfig,
    ):
        self.client = client
        self.default_bucket_name = default_bucket_name
        self.config = config

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *exc: Any) -> None:
        self.client.close()

    def read(
        self,
        fp: SecureFilePathLocation,
        type_: type | None,
        bucket_name: str | None = None,
    ) -> BlobRetrieval:
        if bucket_name is None:
            bucket_name = self.default_bucket_name
        # this will generate the url, the SecureFilePathLocation also handles the logic
        # that decides whether to use a direct connection to azure/aws/gcp or via seaweed
        return fp.generate_url(self, type_, bucket_name)

    def allocate(
        self, obj: CreateBlobStorageEntry
    ) -> SecureFilePathLocation | SyftError:
        try:
            file_name = obj.file_name
            result = self.client.create_multipart_upload(
                Bucket=self.default_bucket_name,
                Key=file_name,
            )
            upload_id = result["UploadId"]
            return SeaweedSecureFilePathLocation(upload_id=upload_id, path=file_name)
        except BotoClientError as e:
            return SyftError(
                message=f"Failed to allocate space for {obj} with error: {e}"
            )

    def write(self, obj: BlobStorageEntry) -> BlobDeposit:
        total_parts = math.ceil(obj.file_size / DEFAULT_FILE_PART_SIZE)

        urls = [
            GridURL.from_url(
                self.client.generate_presigned_url(
                    ClientMethod="upload_part",
                    Params={
                        "Bucket": self.default_bucket_name,
                        "Key": obj.location.path,
                        "UploadId": obj.location.upload_id,
                        "PartNumber": i + 1,
                    },
                    ExpiresIn=WRITE_EXPIRATION_TIME,
                )
            )
            for i in range(total_parts)
        ]
        return SeaweedFSBlobDeposit(
            blob_storage_entry_id=obj.id, urls=urls, size=obj.file_size
        )

    def complete_multipart_upload(
        self,
        blob_entry: BlobStorageEntry,
        etags: list,
    ) -> SyftError | SyftSuccess:
        try:
            self.client.complete_multipart_upload(
                Bucket=self.default_bucket_name,
                Key=blob_entry.location.path,
                MultipartUpload={"Parts": etags},
                UploadId=blob_entry.location.upload_id,
            )
            return SyftSuccess(message="Successfully saved file.")
        except BotoClientError as e:
            return SyftError(message=str(e))

    def delete(
        self,
        fp: SecureFilePathLocation,
    ) -> SyftSuccess | SyftError:
        try:
            self.client.delete_object(Bucket=self.default_bucket_name, Key=fp.path)
            return SyftSuccess(message="Successfully deleted file.")
        except BotoClientError as e:
            return SyftError(message=str(e))


@serializable()
class SeaweedFSConfig(BlobStorageConfig):
    client_type: type[BlobStorageClient] = SeaweedFSClient
    client_config: SeaweedFSClientConfig



# stdlib
from datetime import datetime
from datetime import timedelta
from enum import Enum
from typing import Any

# third party
from pydantic import field_validator
from pydantic import model_validator
from result import Err
from result import Ok
from result import Result
from typing_extensions import Self

# relative
from ...client.api import APIRegistry
from ...client.api import SyftAPICall
from ...node.credentials import SyftVerifyKey
from ...serde.serializable import serializable
from ...service.queue.queue_stash import QueueItem
from ...service.worker.worker_pool import SyftWorker
from ...store.document_store import BaseStash
from ...store.document_store import DocumentStore
from ...store.document_store import PartitionKey
from ...store.document_store import PartitionSettings
from ...store.document_store import QueryKeys
from ...store.document_store import UIDPartitionKey
from ...types.datetime import DateTime
from ...types.syft_object import SYFT_OBJECT_VERSION_2
from ...types.syft_object import SYFT_OBJECT_VERSION_4
from ...types.syft_object import SyftObject
from ...types.syft_object import short_uid
from ...types.syncable_object import SyncableSyftObject
from ...types.uid import UID
from ...util import options
from ...util.colors import SURFACE
from ...util.markdown import as_markdown_code
from ...util.telemetry import instrument
from ..action.action_data_empty import ActionDataLink
from ..action.action_object import Action
from ..action.action_object import ActionObject
from ..action.action_permissions import ActionObjectPermission
from ..response import SyftError
from ..response import SyftNotReady
from ..response import SyftSuccess
from ..user.user import UserView


@serializable()
class JobStatus(str, Enum):
    CREATED = "created"
    PROCESSING = "processing"
    ERRORED = "errored"
    COMPLETED = "completed"
    INTERRUPTED = "interrupted"


@serializable()
class Job(SyncableSyftObject):
    __canonical_name__ = "JobItem"
    __version__ = SYFT_OBJECT_VERSION_4

    id: UID
    node_uid: UID
    result: Any | None = None
    resolved: bool = False
    status: JobStatus = JobStatus.CREATED
    log_id: UID | None = None
    parent_job_id: UID | None = None
    n_iters: int | None = 0
    current_iter: int | None = None
    creation_time: str | None = None
    action: Action | None = None
    job_pid: int | None = None
    job_worker_id: UID | None = None
    updated_at: DateTime | None = None
    user_code_id: UID | None = None

    __attr_searchable__ = ["parent_job_id", "job_worker_id", "status", "user_code_id"]
    __repr_attrs__ = ["id", "result", "resolved", "progress", "creation_time"]
    __exclude_sync_diff_attrs__ = ["action"]

    @field_validator("creation_time")
    @classmethod
    def check_time(cls, time: Any) -> Any:
        return str(datetime.now()) if time is None else time

    @model_validator(mode="after")
    def check_user_code_id(self) -> Self:
        if self.action is not None:
            if self.user_code_id is None:
                self.user_code_id = self.action.user_code_id
            elif self.action.user_code_id != self.user_code_id:
                raise ValueError(
                    "user_code_id does not match the action's user_code_id",
                    self.__class__,
                )

        return self

    @property
    def action_display_name(self) -> str:
        if self.action is None:
            return "action"
        else:
            # hacky
            self.action.syft_node_location = self.syft_node_location
            self.action.syft_client_verify_key = self.syft_client_verify_key
            return self.action.job_display_name

    @property
    def time_remaining_string(self) -> str | None:
        # update state
        self.fetch()
        if (
            self.current_iter is not None
            and self.n_iters is not None
            and self.n_iters != 0
        ):
            percentage = round((self.current_iter / self.n_iters) * 100)
            blocks_filled = round(percentage / 20)
            blocks_empty = 5 - blocks_filled
            blocks_filled_str = "█" * blocks_filled
            blocks_empty_str = "&nbsp;&nbsp;" * blocks_empty
            return f"{percentage}% |{blocks_filled_str}{blocks_empty_str}|\n{self.current_iter}/{self.n_iters}\n"
        return None

    @property
    def worker(self) -> SyftWorker | SyftError:
        api = APIRegistry.api_for(
            node_uid=self.syft_node_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            return SyftError(
                message=f"Can't access Syft API. You must login to {self.syft_node_location}"
            )
        return api.services.worker.get(self.job_worker_id)

    @property
    def eta_string(self) -> str | None:
        if (
            self.current_iter is None
            or self.current_iter == 0
            or self.n_iters is None
            or self.creation_time is None
        ):
            return None

        def format_timedelta(local_timedelta: timedelta) -> str:
            total_seconds = int(local_timedelta.total_seconds())
            hours, leftover = divmod(total_seconds, 3600)
            minutes, seconds = divmod(leftover, 60)

            hours_string = f"{hours}:" if hours != 0 else ""
            minutes_string = f"{minutes}:".zfill(3)
            seconds_string = f"{seconds}".zfill(2)

            return f"{hours_string}{minutes_string}{seconds_string}"

        now = datetime.now()
        time_passed = now - datetime.fromisoformat(self.creation_time)
        iter_duration_seconds: float = time_passed.total_seconds() / self.current_iter
        iters_remaining = self.n_iters - self.current_iter

        # TODO: Adjust by the number of consumers
        time_remaining = timedelta(seconds=iters_remaining * iter_duration_seconds)
        time_passed_str = format_timedelta(time_passed)
        time_remaining_str = format_timedelta(time_remaining)

        if iter_duration_seconds >= 1:
            iter_duration: timedelta = timedelta(seconds=iter_duration_seconds)
            iter_duration_str = f"{format_timedelta(iter_duration)}s/it"
        else:
            iters_per_second = round(1 / iter_duration_seconds)
            iter_duration_str = f"{iters_per_second}it/s"

        return f"[{time_passed_str}<{time_remaining_str}]\n{iter_duration_str}"

    @property
    def progress(self) -> str | None:
        if self.status in [JobStatus.PROCESSING, JobStatus.COMPLETED]:
            if self.current_iter is None:
                return ""
            else:
                if self.n_iters is not None:
                    return self.time_remaining_string
                # if self.current_iter !=0
                # we can compute the remaining time

                # we cannot compute the remaining time
                else:
                    n_iters_str = "?" if self.n_iters is None else str(self.n_iters)
                    return f"{self.current_iter}/{n_iters_str}"
        else:
            return ""

    def info(
        self,
        public_metadata: bool = True,
        result: bool = False,
    ) -> "JobInfo":
        return JobInfo.from_job(self, public_metadata, result)

    def apply_info(self, info: "JobInfo") -> None:
        if info.includes_metadata:
            for attr in info.__public_metadata_attrs__:
                setattr(self, attr, getattr(info, attr))

        if info.includes_result:
            self.result = info.result

    def restart(self, kill: bool = False) -> None:
        if kill:
            self.kill()
        self.fetch()
        if not self.has_parent:
            # this is currently the limitation, we will need to implement
            # killing toplevel jobs later
            print("Can only kill nested jobs")
        elif kill or (
            self.status != JobStatus.PROCESSING and self.status != JobStatus.CREATED
        ):
            api = APIRegistry.api_for(
                node_uid=self.syft_node_location,
                user_verify_key=self.syft_client_verify_key,
            )
            if api is None:
                raise ValueError(
                    f"Can't access Syft API. You must login to {self.syft_node_location}"
                )
            call = SyftAPICall(
                node_uid=self.node_uid,
                path="job.restart",
                args=[],
                kwargs={"uid": self.id},
                blocking=True,
            )

            api.make_call(call)
        else:
            print(
                "Job is running or scheduled, if you want to kill it use job.kill() first"
            )
        return None

    def kill(self) -> SyftError | None:
        if self.job_pid is not None:
            api = APIRegistry.api_for(
                node_uid=self.syft_node_location,
                user_verify_key=self.syft_client_verify_key,
            )
            if api is None:
                return SyftError(
                    message=f"Can't access Syft API. You must login to {self.syft_node_location}"
                )
            call = SyftAPICall(
                node_uid=self.node_uid,
                path="job.kill",
                args=[],
                kwargs={"id": self.id},
                blocking=True,
            )
            api.make_call(call)
            return None
        else:
            return SyftError(
                message="Job is not running or isn't running in multiprocessing mode."
            )

    def fetch(self) -> None:
        api = APIRegistry.api_for(
            node_uid=self.syft_node_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            raise ValueError(
                f"Can't access Syft API. You must login to {self.syft_node_location}"
            )
        call = SyftAPICall(
            node_uid=self.node_uid,
            path="job.get",
            args=[],
            kwargs={"uid": self.id},
            blocking=True,
        )
        job: Job = api.make_call(call)
        self.resolved = job.resolved
        if job.resolved:
            self.result = job.result

        self.status = job.status
        self.n_iters = job.n_iters
        self.current_iter = job.current_iter

    @property
    def subjobs(self) -> list[QueueItem] | SyftError:
        api = APIRegistry.api_for(
            node_uid=self.syft_node_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            return SyftError(
                message=f"Can't access Syft API. You must login to {self.syft_node_location}"
            )
        return api.services.job.get_subjobs(self.id)

    @property
    def owner(self) -> UserView | SyftError:
        api = APIRegistry.api_for(
            node_uid=self.syft_node_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            return SyftError(
                message=f"Can't access Syft API. You must login to {self.syft_node_location}"
            )
        return api.services.user.get_current_user(self.id)

    def _get_log_objs(self) -> SyftObject | SyftError:
        api = APIRegistry.api_for(
            node_uid=self.node_uid,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            raise ValueError(f"api is None. You must login to {self.node_uid}")
        return api.services.log.get(self.log_id)

    def logs(
        self, stdout: bool = True, stderr: bool = True, _print: bool = True
    ) -> str | None:
        api = APIRegistry.api_for(
            node_uid=self.syft_node_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            return f"Can't access Syft API. You must login to {self.syft_node_location}"
        results = []
        if stdout:
            stdout_log = api.services.log.get_stdout(self.log_id)
            if isinstance(stdout_log, SyftError):
                results.append(f"Log {self.log_id} not available")
            else:
                results.append(stdout_log)

        if stderr:
            try:
                std_err_log = api.services.log.get_error(self.log_id)
                results.append(std_err_log)
            except Exception:
                # no access
                if isinstance(self.result, Err):
                    results.append(self.result.value)
        else:
            # add short error
            if isinstance(self.result, Err):
                results.append(self.result.value)

        results_str = "\n".join(results)
        if not _print:
            return results_str
        else:
            print(results_str)
            return None

    # def __repr__(self) -> str:
    #     return f"<Job: {self.id}>: {self.status}"

    def _coll_repr_(self) -> dict[str, Any]:
        logs = self.logs(_print=False, stderr=False)
        if logs is not None:
            log_lines = logs.split("\n")
        subjobs = self.subjobs
        if len(log_lines) > 2:
            logs = f"... ({len(log_lines)} lines)\n" + "\n".join(log_lines[-2:])

        created_time = self.creation_time[:-7] if self.creation_time is not None else ""
        return {
            "status": f"{self.action_display_name}: {self.status}"
            + (
                f"\non worker {short_uid(self.job_worker_id)}"
                if self.job_worker_id
                else ""
            ),
            "progress": self.progress,
            "eta": self.eta_string,
            "created": f"{created_time} by {self.owner.email}",
            "logs": logs,
            # "result": result,
            # "parent_id": str(self.parent_job_id) if self.parent_job_id else "-",
            "subjobs": len(subjobs),
        }

    @property
    def has_parent(self) -> bool:
        return self.parent_job_id is not None

    def _repr_markdown_(self, wrap_as_python: bool = True, indent: int = 0) -> str:
        _ = self.resolve
        logs = self.logs(_print=False)
        if logs is not None:
            logs_w_linenr = "\n".join(
                [f"{i} {line}" for i, line in enumerate(logs.rstrip().split("\n"))]
            )

        if self.status == JobStatus.COMPLETED:
            logs_w_linenr += "\nJOB COMPLETED"

        md = f"""class Job:
    id: UID = {self.id}
    status: {self.status}
    has_parent: {self.has_parent}
    result: {self.result.__str__()}
    logs:

{logs_w_linenr}
    """
        return as_markdown_code(md)

    def wait(
        self, job_only: bool = False, timeout: int | None = None
    ) -> Any | SyftNotReady:
        # stdlib
        from time import sleep

        api = APIRegistry.api_for(
            node_uid=self.syft_node_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if self.resolved:
            return self.resolve

        if not job_only and self.result is not None:
            self.result.wait()

        if api is None:
            raise ValueError(
                f"Can't access Syft API. You must login to {self.syft_node_location}"
            )
        print_warning = True
        counter = 0
        while True:
            self.fetch()
            if print_warning and self.result is not None:
                result_obj = api.services.action.get(
                    self.result.id, resolve_nested=False
                )
                if isinstance(result_obj.syft_action_data, ActionDataLink) and job_only:
                    print(
                        "You're trying to wait on a job that has a link as a result."
                        "This means that the job may be ready but the linked result may not."
                        "Use job.wait().get() instead to wait for the linked result."
                    )
                    print_warning = False
            sleep(1)
            if self.resolved:
                break  # type: ignore[unreachable]
            # TODO: fix the mypy issue
            if timeout is not None:
                counter += 1
                if counter > timeout:
                    return SyftError(message="Reached Timeout!")
        return self.resolve  # type: ignore[unreachable]

    @property
    def resolve(self) -> Any | SyftNotReady:
        if not self.resolved:
            self.fetch()

        if self.resolved:
            return self.result
        return SyftNotReady(message=f"{self.id} not ready yet.")

    def get_sync_dependencies(self, **kwargs: dict) -> list[UID]:  # type: ignore
        dependencies = []
        if self.result is not None:
            dependencies.append(self.result.id.id)

        if self.log_id:
            dependencies.append(self.log_id)

        subjob_ids = [subjob.id for subjob in self.subjobs]
        dependencies.extend(subjob_ids)

        if self.user_code_id is not None:
            dependencies.append(self.user_code_id)

        return dependencies


@serializable()
class JobInfo(SyftObject):
    __canonical_name__ = "JobInfo"
    __version__ = SYFT_OBJECT_VERSION_2

    __repr_attrs__ = [
        "resolved",
        "status",
        "n_iters",
        "current_iter",
        "creation_time",
    ]
    __public_metadata_attrs__ = [
        "resolved",
        "status",
        "n_iters",
        "current_iter",
        "creation_time",
    ]
    # Separate check if the job has logs, result, or metadata
    # None check is not enough because the values we set could be None
    includes_metadata: bool
    includes_result: bool
    # TODO add logs (error reporting PRD)

    resolved: bool | None = None
    status: JobStatus | None = None
    n_iters: int | None = None
    current_iter: int | None = None
    creation_time: str | None = None

    result: ActionObject | None = None

    def _repr_html_(self) -> str:
        metadata_str = ""
        if self.includes_metadata:
            metadata_str += "<h4>Public metadata</h4>"
            for attr in self.__public_metadata_attrs__:
                value = getattr(self, attr, None)
                if value is not None:
                    metadata_str += f"<p style='margin-left: 10px;'><strong>{attr}:</strong> {value}</p>"

        result_str = "<h4>Result</h4>"
        if self.includes_result:
            result_str += f"<p style='margin-left: 10px;'>{str(self.result)}</p>"
        else:
            result_str += "<p style='margin-left: 10px;'><i>No result included</i></p>"

        return f"""
            <style>
            .job-info {{color: {SURFACE[options.color_theme]};}}
            </style>
            <div class='job-info'>
                <h3>JobInfo</h3>
                {metadata_str}
                {result_str}
            </div>
        """

    @classmethod
    def from_job(
        cls,
        job: Job,
        metadata: bool = False,
        result: bool = False,
    ) -> Self:
        info = cls(
            includes_metadata=metadata,
            includes_result=result,
        )

        if metadata:
            for attr in cls.__public_metadata_attrs__:
                setattr(info, attr, getattr(job, attr))

        if result:
            if not job.resolved:
                raise ValueError("Cannot sync result of unresolved job")
            if not isinstance(job.result, ActionObject):
                raise ValueError("Could not sync result of job")
            info.result = job.result

        return info


@instrument
@serializable()
class JobStash(BaseStash):
    object_type = Job
    settings: PartitionSettings = PartitionSettings(
        name=Job.__canonical_name__, object_type=Job
    )

    def __init__(self, store: DocumentStore) -> None:
        super().__init__(store=store)

    def set_result(
        self,
        credentials: SyftVerifyKey,
        item: Job,
        add_permissions: list[ActionObjectPermission] | None = None,
    ) -> Result[Job | None, str]:
        valid = self.check_type(item, self.object_type)
        if valid.is_err():
            return SyftError(message=valid.err())
        return super().update(credentials, item, add_permissions)

    def set_placeholder(
        self,
        credentials: SyftVerifyKey,
        item: Job,
        add_permissions: list[ActionObjectPermission] | None = None,
    ) -> Result[Job, str]:
        # 🟡 TODO 36: Needs distributed lock
        if not item.resolved:
            exists = self.get_by_uid(credentials, item.id)
            if exists.is_ok() and exists.ok() is None:
                valid = self.check_type(item, self.object_type)
                if valid.is_err():
                    return SyftError(message=valid.err())
                return super().set(credentials, item, add_permissions)
        return item

    def get_by_uid(
        self, credentials: SyftVerifyKey, uid: UID
    ) -> Result[Job | None, str]:
        qks = QueryKeys(qks=[UIDPartitionKey.with_obj(uid)])
        item = self.query_one(credentials=credentials, qks=qks)
        return item

    def get_by_parent_id(
        self, credentials: SyftVerifyKey, uid: UID
    ) -> Result[Job | None, str]:
        qks = QueryKeys(
            qks=[PartitionKey(key="parent_job_id", type_=UID).with_obj(uid)]
        )
        item = self.query_all(credentials=credentials, qks=qks)
        return item

    def delete_by_uid(
        self, credentials: SyftVerifyKey, uid: UID
    ) -> Result[SyftSuccess, str]:
        qk = UIDPartitionKey.with_obj(uid)
        result = super().delete(credentials=credentials, qk=qk)
        if result.is_ok():
            return Ok(SyftSuccess(message=f"ID: {uid} deleted"))
        return result

    def get_active(self, credentials: SyftVerifyKey) -> Result[SyftSuccess, str]:
        qks = QueryKeys(
            qks=[
                PartitionKey(key="status", type_=JobStatus).with_obj(
                    JobStatus.PROCESSING
                )
            ]
        )
        return self.query_all(credentials=credentials, qks=qks)

    def get_by_worker(
        self, credentials: SyftVerifyKey, worker_id: str
    ) -> Result[list[Job], str]:
        qks = QueryKeys(
            qks=[PartitionKey(key="job_worker_id", type_=str).with_obj(worker_id)]
        )
        return self.query_all(credentials=credentials, qks=qks)

    def get_by_user_code_id(
        self, credentials: SyftVerifyKey, user_code_id: UID
    ) -> Result[list[Job], str]:
        qks = QueryKeys(
            qks=[PartitionKey(key="user_code_id", type_=UID).with_obj(user_code_id)]
        )

        return self.query_all(credentials=credentials, qks=qks)
